# original code from package: gameobjects
# Home-page: http://code.google.com/p/gameobjects/
# Author: Will McGugan
# Download-URL: http://code.google.com/p/gameobjects/downloads/list
# Copyright (c) 2010-2021 Manfred Moitzi
# License: MIT License
from typing import Sequence, Iterable, List, Tuple, TYPE_CHECKING, Iterator
import math
from math import sin, cos, tan
from itertools import chain

# The pure Python implementation can't import from ._ctypes or ezdxf.math!
from ._vector import Vec3, X_AXIS, Y_AXIS, Z_AXIS, NULLVEC

if TYPE_CHECKING:
    from ezdxf.eztypes import Vertex

__all__ = ["Matrix44"]


# removed array.array because array is optimized for space not speed, and space
# optimization is not needed


def floats(items: Iterable) -> List[float]:
    return [float(v) for v in items]


class Matrix44:
    """This is a pure Python implementation for 4x4 `transformation matrices`_ ,
    to avoid dependency to big numerical packages like :mod:`numpy`, before binary
    wheels, installation of these packages wasn't always easy on Windows.

    The utility functions for constructing transformations and transforming
    vectors and points assumes that vectors are stored as row vectors, meaning
    when multiplied, transformations are applied left to right (e.g. vAB
    transforms v by A then by B).

    Matrix44 initialization:

        - ``Matrix44()`` returns the identity matrix.
        - ``Matrix44(values)`` values is an iterable with the 16 components of
          the matrix.
        - ``Matrix44(row1, row2, row3, row4)`` four rows, each row with four
          values.

    .. _transformation matrices: https://en.wikipedia.org/wiki/Transformation_matrix

    """

    # fmt: off
    _identity = (
        1.0, 0.0, 0.0, 0.0,
        0.0, 1.0, 0.0, 0.0,
        0.0, 0.0, 1.0, 0.0,
        0.0, 0.0, 0.0, 1.0
    )
    # fmt: on
    __slots__ = ("_matrix",)

    def __init__(self, *args):
        """
        Matrix44() is the identity matrix.

        Matrix44(values) values is an iterable with the 16 components of the matrix.

        Matrix44(row1, row2, row3, row4) four rows, each row with four values.

        """
        nargs = len(args)
        if nargs == 0:
            self._matrix = floats(Matrix44._identity)
        elif nargs == 1:
            self._matrix = floats(args[0])
        elif nargs == 4:
            self._matrix = floats(chain(*args))
        else:
            raise ValueError(
                "Invalid count of arguments (4 row vectors or one "
                "list with 16 values)."
            )
        if len(self._matrix) != 16:
            raise ValueError("Invalid matrix count")

    def __repr__(self) -> str:
        """Returns the representation string of the matrix:
        ``Matrix44((col0, col1, col2, col3), (...), (...), (...))``
        """

        def format_row(row):
            return "(%s)" % ", ".join(str(value) for value in row)

        return "Matrix44(%s)" % ", ".join(
            format_row(row) for row in self.rows()
        )

    def get_2d_transformation(self) -> Tuple[float, ...]:
        """Returns a the 2D transformation as a row-major matrix in a linear
        array (tuple).

        A more correct transformation could be implemented like so:
        https://stackoverflow.com/questions/10629737/convert-3d-4x4-rotation-matrix-into-2d
        """
        m = self._matrix
        return m[0], m[1], 0.0, m[4], m[5], 0.0, m[12], m[13], 1.0

    def get_row(self, row: int) -> Tuple[float, ...]:
        """Get row as list of of four float values.

        Args:
            row: row index [0 .. 3]

        """
        if 0 <= row < 4:
            index = row * 4
            return tuple(self._matrix[index: index + 4])
        else:
            raise IndexError(f"invalid row index: {row}")

    def set_row(self, row: int, values: Sequence[float]) -> None:
        """
        Sets the values in a row.

        Args:
            row: row index [0 .. 3]
            values: iterable of four row values

        """
        if 0 <= row < 4:
            index = row * 4
            self._matrix[index: index + len(values)] = floats(values)
        else:
            raise IndexError(f"invalid row index: {row}")

    def get_col(self, col: int) -> Tuple[float, ...]:
        """
        Returns a column as a tuple of four floats.

        Args:
            col: column index [0 .. 3]
        """
        if 0 <= col < 4:
            m = self._matrix
            return m[col], m[col + 4], m[col + 8], m[col + 12]
        else:
            raise IndexError(f"invalid row index: {col}")

    def set_col(self, col: int, values: Sequence[float]):
        """
        Sets the values in a column.

        Args:
            col: column index [0 .. 3]
            values: iterable of four column values

        """
        if 0 <= col < 4:
            m = self._matrix
            a, b, c, d = values
            m[col] = float(a)
            m[col + 4] = float(b)
            m[col + 8] = float(c)
            m[col + 12] = float(d)
        else:
            raise IndexError(f"invalid row index: {col}")

    def copy(self) -> "Matrix44":
        """Returns a copy of same type."""
        return self.__class__(self._matrix)

    __copy__ = copy

    @property
    def origin(self) -> Vec3:
        m = self._matrix
        return Vec3(m[12], m[13], m[14])

    @origin.setter
    def origin(self, v: "Vertex") -> None:
        m = self._matrix
        m[12], m[13], m[14] = Vec3(v)

    @property
    def ux(self) -> Vec3:
        return Vec3(self._matrix[0:3])

    @property
    def uy(self) -> Vec3:
        return Vec3(self._matrix[4:7])

    @property
    def uz(self) -> Vec3:
        return Vec3(self._matrix[8:11])

    @property
    def is_cartesian(self) -> bool:
        """Returns ``True`` if target coordinate system is a right handed
        orthogonal coordinate system.
        """
        return self.uy.cross(self.uz).normalize().isclose(self.ux.normalize())

    @property
    def is_orthogonal(self) -> bool:
        """Returns ``True`` if target coordinate system has orthogonal axis.

        Does not check for left- or right handed orientation, any orientation
        of the axis valid.

        """
        ux = self.ux.normalize()
        uy = self.uy.normalize()
        uz = self.uz.normalize()
        return (
            abs(ux.dot(uy)) <= 1e-9
            and abs(ux.dot(uz)) <= 1e-9
            and abs(uy.dot(uz)) <= 1e-9
        )

    @classmethod
    def scale(cls, sx: float, sy: float = None, sz: float = None) -> "Matrix44":
        """Returns a scaling transformation matrix. If `sy` is ``None``,
        `sy` = `sx`, and if `sz` is ``None`` `sz` = `sx`.

        """
        if sy is None:
            sy = sx
        if sz is None:
            sz = sx
        # fmt: off
        m = cls([
            float(sx), 0., 0., 0.,
            0., float(sy), 0., 0.,
            0., 0., float(sz), 0.,
            0., 0., 0., 1.
        ])
        # fmt: on
        return m

    @classmethod
    def translate(cls, dx: float, dy: float, dz: float) -> "Matrix44":
        """Returns a translation matrix for translation vector (dx, dy, dz)."""
        # fmt: off
        return cls([
            1., 0., 0., 0.,
            0., 1., 0., 0.,
            0., 0., 1., 0.,
            float(dx), float(dy), float(dz), 1.
        ])
        # fmt: on

    @classmethod
    def x_rotate(cls, angle: float) -> "Matrix44":
        """Returns a rotation matrix about the x-axis.

        Args:
            angle: rotation angle in radians

        """
        cos_a = cos(angle)
        sin_a = sin(angle)
        # fmt: off
        return cls([
            1., 0., 0., 0.,
            0., cos_a, sin_a, 0.,
            0., -sin_a, cos_a, 0.,
            0., 0., 0., 1.
        ])
        # fmt: on

    @classmethod
    def y_rotate(cls, angle: float) -> "Matrix44":
        """Returns a rotation matrix about the y-axis.

        Args:
            angle: rotation angle in radians

        """
        cos_a = cos(angle)
        sin_a = sin(angle)
        # fmt: off
        return cls([
            cos_a, 0., -sin_a, 0.,
            0., 1., 0., 0.,
            sin_a, 0., cos_a, 0.,
            0., 0., 0., 1.
        ])
        # fmt: on

    @classmethod
    def z_rotate(cls, angle: float) -> "Matrix44":
        """Returns a rotation matrix about the z-axis.

        Args:
            angle: rotation angle in radians

        """
        cos_a = cos(angle)
        sin_a = sin(angle)
        # fmt: off
        return cls([
            cos_a, sin_a, 0., 0.,
            -sin_a, cos_a, 0., 0.,
            0., 0., 1., 0.,
            0., 0., 0., 1.
        ])
        # fmt: on

    @classmethod
    def axis_rotate(cls, axis: "Vertex", angle: float) -> "Matrix44":
        """Returns a rotation matrix about an arbitrary `axis`.

        Args:
            axis: rotation axis as ``(x, y, z)`` tuple or :class:`Vec3` object
            angle: rotation angle in radians

        """
        c = cos(angle)
        s = sin(angle)
        omc = 1.0 - c
        x, y, z = Vec3(axis).normalize()
        # fmt: off
        return cls([
            x * x * omc + c, y * x * omc + z * s, x * z * omc - y * s, 0.,
            x * y * omc - z * s, y * y * omc + c, y * z * omc + x * s, 0.,
            x * z * omc + y * s, y * z * omc - x * s, z * z * omc + c, 0.,
            0., 0., 0., 1.
        ])
        # fmt: on

    @classmethod
    def xyz_rotate(
        cls, angle_x: float, angle_y: float, angle_z: float
    ) -> "Matrix44":
        """
        Returns a rotation matrix for rotation about each axis.

        Args:
            angle_x: rotation angle about x-axis in radians
            angle_y: rotation angle about y-axis in radians
            angle_z: rotation angle about z-axis in radians

        """
        cx = cos(angle_x)
        sx = sin(angle_x)
        cy = cos(angle_y)
        sy = sin(angle_y)
        cz = cos(angle_z)
        sz = sin(angle_z)

        sxsy = sx * sy
        cxsy = cx * sy
        # fmt: off
        return cls([
            cy * cz, sxsy * cz + cx * sz, -cxsy * cz + sx * sz, 0.,
            -cy * sz, -sxsy * sz + cx * cz, cxsy * sz + sx * cz, 0.,
            sy, -sx * cy, cx * cy, 0.,
            0., 0., 0., 1.
        ])
        # fmt: on

    @classmethod
    def shear_xy(cls, angle_x: float = 0, angle_y: float = 0) -> "Matrix44":
        """Returns a translation matrix for shear mapping (visually similar
        to slanting) in the xy-plane.

        Args:
            angle_x: slanting angle in x direction in radians
            angle_y: slanting angle in y direction in radians

        """
        tx = math.tan(angle_x)
        ty = math.tan(angle_y)
        # fmt: off
        return cls([
            1., ty, 0., 0.,
            tx, 1., 0., 0.,
            0., 0., 1., 0.,
            0., 0., 0., 1.
        ])
        # fmt: on

    @classmethod
    def perspective_projection(
        cls,
        left: float,
        right: float,
        top: float,
        bottom: float,
        near: float,
        far: float,
    ) -> "Matrix44":
        """Returns a matrix for a 2D projection.

        Args:
            left: Coordinate of left of screen
            right: Coordinate of right of screen
            top: Coordinate of the top of the screen
            bottom: Coordinate of the bottom of the screen
            near: Coordinate of the near clipping plane
            far: Coordinate of the far clipping plane

        """
        # fmt: off
        return cls([
            (2. * near) / (right - left), 0., 0., 0.,
            0., (2. * near) / (top - bottom), 0., 0.,
            (right + left) / (right - left), (top + bottom) / (top - bottom),
            -((far + near) / (far - near)), -1.,
            0., 0., -((2. * far * near) / (far - near)), 0.
        ])
        # fmt: on

    @classmethod
    def perspective_projection_fov(
        cls, fov: float, aspect: float, near: float, far: float
    ) -> "Matrix44":
        """Returns a matrix for a 2D projection.

        Args:
            fov: The field of view (in radians)
            aspect: The aspect ratio of the screen (width / height)
            near: Coordinate of the near clipping plane
            far: Coordinate of the far clipping plane

        """
        vrange = near * tan(fov / 2.0)
        left = -vrange * aspect
        right = vrange * aspect
        bottom = -vrange
        top = vrange
        return cls.perspective_projection(left, right, bottom, top, near, far)

    @staticmethod
    def chain(*matrices: "Matrix44") -> "Matrix44":
        """Compose a transformation matrix from one or more `matrices`."""
        transformation = Matrix44()
        for matrix in matrices:
            transformation *= matrix
        return transformation

    @staticmethod
    def ucs(
        ux: Vec3 = X_AXIS,
        uy: Vec3 = Y_AXIS,
        uz: Vec3 = Z_AXIS,
        origin: Vec3 = NULLVEC,
    ) -> "Matrix44":
        """Returns a matrix for coordinate transformation from WCS to UCS.
        For transformation from UCS to WCS, transpose the returned matrix.

        Args:
            ux: x-axis for UCS as unit vector
            uy: y-axis for UCS as unit vector
            uz: z-axis for UCS as unit vector
            origin: UCS origin as location vector

        """
        ux_x, ux_y, ux_z = ux
        uy_x, uy_y, uy_z = uy
        uz_x, uz_y, uz_z = uz
        or_x, or_y, or_z = origin
        # fmt: off
        return Matrix44((
            ux_x, ux_y, ux_z, 0,
            uy_x, uy_y, uy_z, 0,
            uz_x, uz_y, uz_z, 0,
            or_x, or_y, or_z, 1,
        ))
        # fmt: on

    def __setitem__(self, index: Tuple[int, int], value: float):
        """Set (row, column) element."""
        row, col = index
        if 0 <= row < 4 and 0 <= col < 4:
            self._matrix[row * 4 + col] = float(value)
        else:
            raise IndexError(f"index out of range: {index}")

    def __getitem__(self, index: Tuple[int, int]):
        """Get (row, column) element."""
        row, col = index
        if 0 <= row < 4 and 0 <= col < 4:
            return self._matrix[row * 4 + col]
        else:
            raise IndexError(f"index out of range: {index}")

    def __iter__(self) -> Iterator[float]:
        """Iterates over all matrix values."""
        return iter(self._matrix)

    def __mul__(self, other: "Matrix44") -> "Matrix44":
        """Returns a new matrix as result of the matrix multiplication with
        another matrix.
        """
        res_matrix = self.copy()
        res_matrix.__imul__(other)
        return res_matrix

    # __matmul__ = __mul__ does not work!

    def __matmul__(self, other: "Matrix44") -> "Matrix44":
        """Returns a new matrix as result of the matrix multiplication with
        another matrix.
        """
        res_matrix = self.copy()
        res_matrix.__imul__(other)
        return res_matrix

    def __imul__(self, other: "Matrix44") -> "Matrix44":
        """Inplace multiplication with another matrix."""
        m1 = self._matrix
        m2 = other._matrix
        # fmt: off
        self._matrix = [
            m1[0] * m2[0] + m1[1] * m2[4] + m1[2] * m2[8] + m1[3] * m2[12],
            m1[0] * m2[1] + m1[1] * m2[5] + m1[2] * m2[9] + m1[3] * m2[13],
            m1[0] * m2[2] + m1[1] * m2[6] + m1[2] * m2[10] + m1[3] * m2[14],
            m1[0] * m2[3] + m1[1] * m2[7] + m1[2] * m2[11] + m1[3] * m2[15],

            m1[4] * m2[0] + m1[5] * m2[4] + m1[6] * m2[8] + m1[7] * m2[12],
            m1[4] * m2[1] + m1[5] * m2[5] + m1[6] * m2[9] + m1[7] * m2[13],
            m1[4] * m2[2] + m1[5] * m2[6] + m1[6] * m2[10] + m1[7] * m2[14],
            m1[4] * m2[3] + m1[5] * m2[7] + m1[6] * m2[11] + m1[7] * m2[15],

            m1[8] * m2[0] + m1[9] * m2[4] + m1[10] * m2[8] + m1[11] * m2[12],
            m1[8] * m2[1] + m1[9] * m2[5] + m1[10] * m2[9] + m1[11] * m2[13],
            m1[8] * m2[2] + m1[9] * m2[6] + m1[10] * m2[10] + m1[11] * m2[14],
            m1[8] * m2[3] + m1[9] * m2[7] + m1[10] * m2[11] + m1[11] * m2[15],

            m1[12] * m2[0] + m1[13] * m2[4] + m1[14] * m2[8] + m1[15] * m2[12],
            m1[12] * m2[1] + m1[13] * m2[5] + m1[14] * m2[9] + m1[15] * m2[13],
            m1[12] * m2[2] + m1[13] * m2[6] + m1[14] * m2[10] + m1[15] * m2[14],
            m1[12] * m2[3] + m1[13] * m2[7] + m1[14] * m2[11] + m1[15] * m2[15]
        ]
        # fmt: on
        return self

    def rows(self) -> Iterable[Tuple[float, ...]]:
        """Iterate over rows as 4-tuples."""
        return (self.get_row(index) for index in (0, 1, 2, 3))

    def columns(self) -> Iterable[Tuple[float, ...]]:
        """Iterate over columns as 4-tuples."""
        return (self.get_col(index) for index in (0, 1, 2, 3))

    def transform(self, vector: "Vertex") -> Vec3:
        """Returns a transformed vertex."""
        m = self._matrix
        x, y, z = Vec3(vector)
        # fmt: off
        return Vec3(
            x * m[0] + y * m[4] + z * m[8] + m[12],
            x * m[1] + y * m[5] + z * m[9] + m[13],
            x * m[2] + y * m[6] + z * m[10] + m[14]
        )
        # fmt: on

    def transform_direction(self, vector: "Vertex", normalize=False) -> Vec3:
        """Returns a transformed direction vector without translation."""
        m = self._matrix
        x, y, z = Vec3(vector)
        # fmt: off
        v = Vec3(
            x * m[0] + y * m[4] + z * m[8],
            x * m[1] + y * m[5] + z * m[9],
            x * m[2] + y * m[6] + z * m[10]
        )
        # fmt: on
        return v.normalize() if normalize else v

    ocs_to_wcs = transform_direction

    def transform_vertices(self, vectors: Iterable["Vertex"]) -> Iterable[Vec3]:
        """Returns an iterable of transformed vertices."""
        # fmt: off
        (
            m0, m1, m2, m3,
            m4, m5, m6, m7,
            m8, m9, m10, m11,
            m12, m13, m14, m15,
        ) = self._matrix
        # fmt: on
        for vector in vectors:
            x, y, z = Vec3(vector)
            # fmt: off
            yield Vec3(
                x * m0 + y * m4 + z * m8 + m12,
                x * m1 + y * m5 + z * m9 + m13,
                x * m2 + y * m6 + z * m10 + m14
            )
            # fmt: on

    def transform_directions(
        self, vectors: Iterable["Vertex"], normalize=False
    ) -> Iterable[Vec3]:
        """Returns an iterable of transformed direction vectors without
        translation.

        """
        m0, m1, m2, m3, m4, m5, m6, m7, m8, m9, m10, *_ = self._matrix
        for vector in vectors:
            x, y, z = Vec3(vector)
            # fmt: off
            v = Vec3(
                x * m0 + y * m4 + z * m8,
                x * m1 + y * m5 + z * m9,
                x * m2 + y * m6 + z * m10
            )
            # fmt: on
            yield v.normalize() if normalize else v

    def ucs_vertex_from_wcs(self, wcs: Vec3) -> Vec3:
        """Returns an UCS vector from WCS vertex.

        Works only if matrix is used as cartesian UCS without scaling.

        (internal API)

        """
        return self.ucs_direction_from_wcs(wcs - self.origin)

    def ucs_direction_from_wcs(self, wcs: Vec3) -> Vec3:
        """Returns UCS direction vector from WCS direction.

        Works only if matrix is used as cartesian UCS without scaling.

        (internal API)

        """
        m0, m1, m2, m3, m4, m5, m6, m7, m8, m9, m10, *_ = self._matrix
        x, y, z = wcs
        # fmt: off
        return Vec3(
            x * m0 + y * m1 + z * m2,
            x * m4 + y * m5 + z * m6,
            x * m8 + y * m9 + z * m10,
        )
        # fmt: on

    ocs_from_wcs = ucs_direction_from_wcs

    def transpose(self) -> None:
        """Swaps the rows for columns inplace."""
        # fmt: off
        (
            m00, m01, m02, m03,
            m10, m11, m12, m13,
            m20, m21, m22, m23,
            m30, m31, m32, m33,
        ) = self._matrix
        self._matrix = [
            m00, m10, m20, m30,
            m01, m11, m21, m31,
            m02, m12, m22, m32,
            m03, m13, m23, m33
        ]
        # fmt: on

    def determinant(self) -> float:
        """Returns determinant."""
        # fmt: off
        (
            m00, m01, m02, m03,
            m10, m11, m12, m13,
            m20, m21, m22, m23,
            m30, m31, m32, m33,
        ) = self._matrix

        return (
            m00 * m11 * m22 * m33 - m00 * m11 * m23 * m32 +
            m00 * m12 * m23 * m31 - m00 * m12 * m21 * m33 +
            m00 * m13 * m21 * m32 - m00 * m13 * m22 * m31 -
            m01 * m12 * m23 * m30 + m01 * m12 * m20 * m33 -
            m01 * m13 * m20 * m32 + m01 * m13 * m22 * m30 -
            m01 * m10 * m22 * m33 + m01 * m10 * m23 * m32 +
            m02 * m13 * m20 * m31 - m02 * m13 * m21 * m30 +
            m02 * m10 * m21 * m33 - m02 * m10 * m23 * m31 +
            m02 * m11 * m23 * m30 - m02 * m11 * m20 * m33 -
            m03 * m10 * m21 * m32 + m03 * m10 * m22 * m31 -
            m03 * m11 * m22 * m30 + m03 * m11 * m20 * m32 -
            m03 * m12 * m20 * m31 + m03 * m12 * m21 * m30
        )
        # fmt: on

    def inverse(self) -> None:
        """Calculates the inverse of the matrix.

        Raises:
             ZeroDivisionError: if matrix has no inverse.

        """
        det = self.determinant()
        f = 1.0 / det  # catch ZeroDivisionError by caller
        # fmt: off
        (
            m00, m01, m02, m03,
            m10, m11, m12, m13,
            m20, m21, m22, m23,
            m30, m31, m32, m33,
        ) = self._matrix
        self._matrix = [
            (
                m12 * m23 * m31 - m13 * m22 * m31 + m13 * m21 * m32 -
                m11 * m23 * m32 - m12 * m21 * m33 + m11 * m22 * m33) * f,
            (
                m03 * m22 * m31 - m02 * m23 * m31 - m03 * m21 * m32 +
                m01 * m23 * m32 + m02 * m21 * m33 - m01 * m22 * m33) * f,
            (
                m02 * m13 * m31 - m03 * m12 * m31 + m03 * m11 * m32 -
                m01 * m13 * m32 - m02 * m11 * m33 + m01 * m12 * m33) * f,
            (
                m03 * m12 * m21 - m02 * m13 * m21 - m03 * m11 * m22 +
                m01 * m13 * m22 + m02 * m11 * m23 - m01 * m12 * m23) * f,
            (
                m13 * m22 * m30 - m12 * m23 * m30 - m13 * m20 * m32 +
                m10 * m23 * m32 + m12 * m20 * m33 - m10 * m22 * m33) * f,
            (
                m02 * m23 * m30 - m03 * m22 * m30 + m03 * m20 * m32 -
                m00 * m23 * m32 - m02 * m20 * m33 + m00 * m22 * m33) * f,
            (
                m03 * m12 * m30 - m02 * m13 * m30 - m03 * m10 * m32 +
                m00 * m13 * m32 + m02 * m10 * m33 - m00 * m12 * m33) * f,
            (
                m02 * m13 * m20 - m03 * m12 * m20 + m03 * m10 * m22 -
                m00 * m13 * m22 - m02 * m10 * m23 + m00 * m12 * m23) * f,
            (
                m11 * m23 * m30 - m13 * m21 * m30 + m13 * m20 * m31 -
                m10 * m23 * m31 - m11 * m20 * m33 + m10 * m21 * m33) * f,
            (
                m03 * m21 * m30 - m01 * m23 * m30 - m03 * m20 * m31 +
                m00 * m23 * m31 + m01 * m20 * m33 - m00 * m21 * m33) * f,
            (
                m01 * m13 * m30 - m03 * m11 * m30 + m03 * m10 * m31 -
                m00 * m13 * m31 - m01 * m10 * m33 + m00 * m11 * m33) * f,
            (
                m03 * m11 * m20 - m01 * m13 * m20 - m03 * m10 * m21 +
                m00 * m13 * m21 + m01 * m10 * m23 - m00 * m11 * m23) * f,
            (
                m12 * m21 * m30 - m11 * m22 * m30 - m12 * m20 * m31 +
                m10 * m22 * m31 + m11 * m20 * m32 - m10 * m21 * m32) * f,
            (
                m01 * m22 * m30 - m02 * m21 * m30 + m02 * m20 * m31 -
                m00 * m22 * m31 - m01 * m20 * m32 + m00 * m21 * m32) * f,
            (
                m02 * m11 * m30 - m01 * m12 * m30 - m02 * m10 * m31 +
                m00 * m12 * m31 + m01 * m10 * m32 - m00 * m11 * m32) * f,
            (
                m01 * m12 * m20 - m02 * m11 * m20 + m02 * m10 * m21 -
                m00 * m12 * m21 - m01 * m10 * m22 + m00 * m11 * m22) * f,
        ]
        # fmt: on



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.const import SUBCLASS_MARKER, DXF2000
from ezdxf.math import NULLVEC, Z_AXIS, X_AXIS
from ezdxf.math.transformtools import transform_extrusion
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import DXFGraphic, acdb_entity
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, DXFNamespace, Matrix44

__all__ = ["Tolerance"]

acdb_tolerance = DefSubclass(
    "AcDbFcf",
    {
        "dimstyle": DXFAttr(
            3,
            default="Standard",
            validator=validator.is_valid_table_name,
        ),
        # Insertion point (in WCS):
        "insert": DXFAttr(10, xtype=XType.point3d, default=NULLVEC),
        # String representing the visual representation of the tolerance:
        "content": DXFAttr(1, default=""),
        "extrusion": DXFAttr(
            210,
            xtype=XType.point3d,
            default=Z_AXIS,
            optional=True,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
        # X-axis direction vector (in WCS):
        "x_axis_vector": DXFAttr(
            11,
            xtype=XType.point3d,
            default=X_AXIS,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
    },
)
acdb_tolerance_group_codes = group_code_mapping(acdb_tolerance)


@register_entity
class Tolerance(DXFGraphic):
    """DXF TOLERANCE entity"""

    DXFTYPE = "TOLERANCE"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_tolerance)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_tolerance_group_codes, subclass=2, recover=True
            )
        return dxf

    def export_entity(self, tagwriter: "TagWriter") -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_tolerance.name)
        self.dxf.export_dxf_attribs(
            tagwriter,
            ["dimstyle", "insert", "content", "extrusion", "x_axis_vector"],
        )

    def transform(self, m: "Matrix44") -> "Tolerance":
        """Transform the TOLERANCE entity by transformation matrix `m` inplace."""
        self.dxf.insert = m.transform(self.dxf.insert)
        self.dxf.x_axis_vector = m.transform_direction(self.dxf.x_axis_vector)
        self.dxf.extrusion, _ = transform_extrusion(self.dxf.extrusion, m)
        self.post_transform(m)
        return self



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
from typing import TYPE_CHECKING, Optional, Iterable
from ezdxf.lldxf import const
from ezdxf.lldxf.tags import Tags
from .dxfentity import SubclassProcessor
from .dxfgfx import DXFGraphic
from . import factory

if TYPE_CHECKING:
    from ezdxf.eztypes import DXFNamespace, TagWriter, DXFEntity


# Group Codes of AcDbProxyEntity
# https://help.autodesk.com/view/OARX/2018/ENU/?guid=GUID-89A690F9-E859-4D57-89EA-750F3FB76C6B
# 100 AcDbProxyEntity
# 90  Proxy entity class ID (always 498)
# 91  Application entity's class ID. Class IDs are based on the order of
#     the class in the CLASSES section. The first class is given the ID of
#     500, the next is 501, and so on
#
# 92  Size of graphics data in bytes < R2010; R2010+ = 160
# 310 Binary graphics data (multiple entries can appear) (optional)
#
# 96  Size of unknown data in bytes < R2010; R2010+ = 162
# 311 Binary entity data (multiple entries can appear) (optional)
#
# 93  Size of entity data in bits <R2010; R2010+ = 161
# 310 Binary entity data (multiple entries can appear) (optional)
#
# 330 or 340 or 350 or 360 - An object ID (multiple entries can appear) (optional)
# 94  0 (indicates end of object ID section)
# 95  Object drawing format when it becomes a proxy (a 32-bit unsigned integer):
#     Low word is AcDbDwgVersion
#     High word is MaintenanceReleaseVersion
# 70  Original custom object data format:
#     0 = DWG format
#     1 = DXF format


@factory.register_entity
class ACADProxyEntity(DXFGraphic):
    """READ ONLY ACAD_PROXY_ENTITY CLASS! DO NOT MODIFY!"""

    DXFTYPE = "ACAD_PROXY_ENTITY"
    MIN_DXF_VERSION_FOR_EXPORT = const.DXF2000

    def __init__(self):
        super().__init__()
        self.acdb_proxy_entity: Optional[Tags] = None

    def copy(self):
        raise const.DXFTypeError(f"Cloning of {self.dxftype()} not supported.")

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor:
            self.acdb_proxy_entity = processor.subclass_by_index(2)
            self.load_proxy_graphic(processor.dxfversion)
        return dxf

    def load_proxy_graphic(self, dxfversion: Optional[str]) -> None:
        if self.acdb_proxy_entity is not None:
            if not dxfversion:
                dxfversion = _detect_dxf_version(self.acdb_proxy_entity)
            length_code = 92 if dxfversion < const.DXF2013 else 160
            self.proxy_graphic = load_proxy_data(
                self.acdb_proxy_entity, length_code, 310
            )

    def export_dxf(self, tagwriter: "TagWriter") -> None:
        # Proxy graphic is stored in AcDbProxyEntity and not as usual in
        # AcDbEntity!
        preserve_proxy_graphic = self.proxy_graphic
        self.proxy_graphic = None
        super().export_dxf(tagwriter)
        self.proxy_graphic = preserve_proxy_graphic

    def export_entity(self, tagwriter: "TagWriter") -> None:
        """Export entity specific data as DXF tags. (internal API)"""
        # Base class and AcDbEntity export is done by parent class
        super().export_entity(tagwriter)
        if self.acdb_proxy_entity is not None:
            tagwriter.write_tags(self.acdb_proxy_entity)
        # XDATA export is done by the parent class

    def __virtual_entities__(self) -> Iterable[DXFGraphic]:
        """Implements the SupportsVirtualEntities protocol. """
        from ezdxf.proxygraphic import ProxyGraphic
        if self.proxy_graphic:
            return ProxyGraphic(self.proxy_graphic, self.doc).virtual_entities()
        return []

    def virtual_entities(self) -> Iterable[DXFGraphic]:
        """Yields proxy graphic as "virtual" entities. """
        return self.__virtual_entities__()


def load_proxy_data(
    tags: Tags, length_code: int, data_code: int = 310
) -> Optional[bytes]:
    try:
        index = tags.tag_index(length_code)
    except const.DXFValueError:
        return None
    binary_data = []
    for code, value in tags[index + 1:]:
        if code == data_code:
            binary_data.append(value)
        else:
            break  # at first tag with group code != data_code
    return b"".join(binary_data)


def _detect_dxf_version(tags: Tags) -> str:
    for tag in tags:
        if 160 <= tag.code < 163:
            return const.DXF2013
    return const.DXF2000



# Copyright (c) 2019-2021, Manfred Moitzi
# License: MIT License
from typing import List, Sequence, TYPE_CHECKING, Iterable
import math
from ezdxf.math import Vec2
from .bbox import BoundingBox2d
from .line import ConstructionLine
from .construct2d import point_to_line_relation

if TYPE_CHECKING:
    from ezdxf.math import Vertex

ABS_TOL = 1e-12

__all__ = ["ConstructionBox"]


class ConstructionBox:
    """
    Helper class to create rectangles.

    Args:
        center: center of rectangle
        width: width of rectangle
        height: height of rectangle
        angle: angle of rectangle in degrees

    """

    def __init__(
        self,
        center: "Vertex" = (0, 0),
        width: float = 1,
        height: float = 1,
        angle: float = 0,
    ):
        self._center = Vec2(center)
        self._width: float = abs(width)
        self._height: float = abs(height)
        self._angle: float = angle  # in degrees
        self._corners: Sequence[Vec2] = tuple()
        self._tainted: bool = True

    @classmethod
    def from_points(cls, p1: "Vertex", p2: "Vertex") -> "ConstructionBox":
        """Creates a box from two opposite corners, box sides are parallel to x-
        and y-axis.

        Args:
            p1: first corner as :class:`Vec2` compatible object
            p2: second corner as :class:`Vec2` compatible object

        """
        _p1 = Vec2(p1)
        _p2 = Vec2(p2)
        width: float = abs(_p2.x - _p1.x)
        height: float = abs(_p2.y - _p1.y)
        center: Vec2 = _p1.lerp(_p2)
        return cls(center=center, width=width, height=height)

    def update(self) -> None:
        if not self._tainted:
            return
        center = self.center
        w2 = Vec2.from_deg_angle(self._angle, self._width / 2.0)
        h2 = Vec2.from_deg_angle(self._angle + 90, self._height / 2.0)
        self._corners = (
            center - w2 - h2,  # lower left
            center + w2 - h2,  # lower right
            center + w2 + h2,  # upper right
            center - w2 + h2,  # upper left
        )
        self._tainted = False

    @property
    def bounding_box(self) -> BoundingBox2d:
        """:class:`BoundingBox2d`"""
        return BoundingBox2d(self.corners)

    @property
    def center(self) -> Vec2:
        """box center"""
        return self._center

    @center.setter
    def center(self, c: "Vertex") -> None:
        self._center = Vec2(c)
        self._tainted = True

    @property
    def width(self) -> float:
        """box width"""
        return self._width

    @width.setter
    def width(self, w: float) -> None:
        self._width = abs(w)
        self._tainted = True

    @property
    def height(self) -> float:
        """box height"""
        return self._height

    @height.setter
    def height(self, h: float) -> None:
        self._height = abs(h)
        self._tainted = True

    @property
    def incircle_radius(self) -> float:
        """incircle radius"""
        return min(self._width, self._height) * 0.5

    @property
    def circumcircle_radius(self) -> float:
        """circum circle radius"""
        return math.hypot(self._width, self._height) * 0.5

    @property
    def angle(self) -> float:
        """rotation angle in degrees"""
        return self._angle

    @angle.setter
    def angle(self, a: float) -> None:
        self._angle = a
        self._tainted = True

    @property
    def corners(self) -> Sequence[Vec2]:
        """box corners as sequence of :class:`Vec2` objects."""
        self.update()
        return self._corners

    def __iter__(self) -> Iterable[Vec2]:
        """Iterable of box corners as :class:`Vec2` objects."""
        return iter(self.corners)

    def __getitem__(self, corner) -> Vec2:
        """Get corner by index `corner`, ``list`` like slicing is supported."""
        return self.corners[corner]

    def __repr__(self) -> str:
        """Returns string representation of box as
        ``ConstructionBox(center, width, height, angle)``"""
        return "ConstructionBox({0.center}, {0.width}, {0.height}, {0.angle})".format(
            self
        )

    def translate(self, dx: float, dy: float) -> None:
        """
        Move box about `dx` in x-axis and about `dy` in y-axis.

        Args:
            dx: translation in x-axis
            dy: translation in y-axis

        """
        self.center += Vec2((dx, dy))

    def expand(self, dw: float, dh: float) -> None:
        """Expand box: `dw` expand width, `dh` expand height."""
        self.width += dw
        self.height += dh

    def scale(self, sw: float, sh: float) -> None:
        """Scale box: `sw` scales width, `sh` scales height."""
        self.width *= sw
        self.height *= sh

    def rotate(self, angle: float) -> None:
        """Rotate box by `angle` in degrees."""
        self.angle += angle

    def is_inside(self, point: "Vertex") -> bool:
        """Returns ``True`` if `point` is inside of box."""
        point = Vec2(point)
        delta = self.center - point
        if abs(self.angle) < ABS_TOL:  # fast path for horizontal rectangles
            return abs(delta.x) <= (self._width / 2.0) and abs(delta.y) <= (
                self._height / 2.0
            )
        else:
            distance = delta.magnitude
            if distance > self.circumcircle_radius:
                return False
            elif distance <= self.incircle_radius:
                return True
            else:
                # inside if point is "left of line" of all border lines.
                p1, p2, p3, p4 = self.corners
                return all(
                    (
                        point_to_line_relation(point, a, b) < 1
                        for a, b in [(p1, p2), (p2, p3), (p3, p4), (p4, p1)]
                    )
                )

    def is_any_corner_inside(self, other: "ConstructionBox") -> bool:
        """Returns ``True`` if any corner of `other` box is inside this box."""
        return any(self.is_inside(p) for p in other.corners)

    def is_overlapping(self, other: "ConstructionBox") -> bool:
        """Returns ``True`` if this box and `other` box do overlap."""
        distance = (self.center - other.center).magnitude
        max_distance = self.circumcircle_radius + other.circumcircle_radius
        if distance > max_distance:
            return False
        min_distance = self.incircle_radius + other.incircle_radius
        if distance <= min_distance:
            return True

        if self.is_any_corner_inside(other):
            return True
        if other.is_any_corner_inside(self):
            return True
        # no corner inside of any box, maybe crossing boxes?
        # check intersection of diagonals
        c1, c2, c3, c4 = self.corners
        diag1 = ConstructionLine(c1, c3)
        diag2 = ConstructionLine(c2, c4)

        t1, t2, t3, t4 = other.corners
        test_diag = ConstructionLine(t1, t3)
        if test_diag.has_intersection(diag1) or test_diag.has_intersection(diag2):
            return True
        test_diag = ConstructionLine(t2, t4)
        if test_diag.has_intersection(diag1) or test_diag.has_intersection(diag2):
            return True

        return False

    def border_lines(self) -> Sequence[ConstructionLine]:
        """Returns border lines of box as sequence of :class:`ConstructionLine`.
        """
        p1, p2, p3, p4 = self.corners
        return (
            ConstructionLine(p1, p2),
            ConstructionLine(p2, p3),
            ConstructionLine(p3, p4),
            ConstructionLine(p4, p1),
        )

    def intersect(self, line: ConstructionLine) -> List[Vec2]:
        """Returns 0, 1 or 2 intersection points between `line` and box border
        lines.

        Args:
            line: line to intersect with border lines

        Returns:
            list of intersection points

            =========== ==================================
            list size   Description
            =========== ==================================
            0           no intersection
            1           line touches box at one corner
            2           line intersects with box
            =========== ==================================

        """
        result = set()
        for border_line in self.border_lines():
            p = line.intersect(border_line)
            if p is not None:
                result.add(p)
        return sorted(result)



# Copyright (c) 2019-2021, Manfred Moitzi
# License: MIT-License
from typing import TYPE_CHECKING
from ezdxf.lldxf import validator
from ezdxf.lldxf.const import SUBCLASS_MARKER, DXF2007
from ezdxf.lldxf.attributes import (
    DXFAttributes,
    DefSubclass,
    DXFAttr,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
)
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import acdb_entity, DXFGraphic
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, DXFNamespace, Matrix44

__all__ = ["Light"]

acdb_light = DefSubclass(
    "AcDbLight",
    {
        "version": DXFAttr(90, default=0),
        # Light name
        "name": DXFAttr(1, default=""),
        # Light type:
        # 1 = distant
        # 2 = point
        # 3 = spot
        "type": DXFAttr(
            70,
            default=1,
            validator=validator.is_in_integer_range(1, 4),
            fixer=RETURN_DEFAULT,
        ),
        "status": DXFAttr(
            290,
            default=1,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        "plot_glyph": DXFAttr(
            291,
            default=0,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        "intensity": DXFAttr(40, default=1),
        # Light position
        "location": DXFAttr(10, xtype=XType.point3d),
        # Target location
        "target": DXFAttr(11, xtype=XType.point3d),
        # Attenuation type:
        # 0 = None
        # 1 = Inverse Linear
        # 2 = Inverse Square
        "attenuation_type": DXFAttr(
            72,
            default=2,
            validator=validator.is_in_integer_range(0, 3),
            fixer=RETURN_DEFAULT,
        ),
        "use_attenuation_limits": DXFAttr(
            292,
            default=0,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        "attenuation_start_limits": DXFAttr(41),
        "attenuation_end_limits": DXFAttr(42),
        "hotspot_angle": DXFAttr(50),  # in degrees
        "falloff_angle": DXFAttr(51),  # in degrees
        "cast_shadows": DXFAttr(
            293,
            default=1,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        # Shadow Type:
        # 0 = Ray traced shadows
        # 1 = Shadow maps
        "shadow_type": DXFAttr(
            73,
            default=0,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        "shadow_map_size": DXFAttr(91),
        "shadow_map_softness": DXFAttr(280),
    },
)
acdb_light_group_codes = group_code_mapping(acdb_light)


@register_entity
class Light(DXFGraphic):
    """DXF LIGHT entity"""

    DXFTYPE = "LIGHT"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_light)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2007

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_light_group_codes, 2, recover=True
            )
        return dxf

    def export_entity(self, tagwriter: "TagWriter") -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_light.name)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "version",
                "name",
                "type",
                "status",
                "plot_glyph",
                "intensity",
                "location",
                "target",
                "attenuation_type",
                "use_attenuation_limits",
                "attenuation_start_limits",
                "attenuation_end_limits",
                "hotspot_angle",
                "falloff_angle",
                "cast_shadows",
                "shadow_type",
                "shadow_map_size",
                "shadow_map_softness",
            ],
        )

    def transform(self, m: "Matrix44") -> "Light":
        """Transform the LIGHT entity by transformation matrix `m` inplace."""
        self.dxf.location = m.transform(self.dxf.location)
        self.dxf.target = m.transform(self.dxf.target)
        self.post_transform(m)
        return self



# Purpose: acdsdata section manager
# Copyright (c) 2014-2021, Manfred Moitzi
# License: MIT License
"""
ACDSDATA entities have NO handles, therefore they can not be stored in the
drawing entity database.

section structure (work in progress):
0 <str> SECTION
2 <str> ACDSDATA
70 <int> 2 # flag?
71 <int> 6 # count of following ACDSSCHEMA entities ??? no, just another flag

0 <str> ACDSSCHEMA           # dxftype: schema definition
90 <int> 0                   # schema number 0, 1, 2, 3 ...
1 <str> AcDb3DSolid_ASM_Data # schema name

2 <str> AcDbDs::ID           # subsection name
280 <int> 10                 # subsection type 10 = ???
91 <int> 8                   # data ???

2 <str> ASM_Data             # subsection name
280 <int> 15                 # subsection type
91 <int> 0                   # data ???
101 <str> ACDSRECORD         # data
95 <int> 0
90 <int> 2
...

0 <str> ACDSSCHEMA
90 <int> 1
1 <str> AcDb_Thumbnail_Schema
...

0 <str> ACDSSCHEMA
90 <int> 2
1 <str> AcDbDs::TreatedAsObjectDataSchema
...

0 <str> ACDSSCHEMA
90 <int> 3
1 <str> AcDbDs::LegacySchema
2 <str> AcDbDs::Legacy
280 <int> 1
91 <int> 0

0 <str> ACDSSCHEMA
90 <int> 4
1 <str> AcDbDs::IndexedPropertySchema
2 <str> AcDs:Indexable
280 <int> 1
91 <int> 0

0 <str> ACDSSCHEMA
90 <int> 5
1 <str> AcDbDs::HandleAttributeSchema
2 <str> AcDbDs::HandleAttribute
280 <int> 7
91 <int> 1
284 <int> 1

0 <str> ACDSRECORD               # dxftype: data record
90 <int> 0                       # ??? flag
2 <str> AcDbDs::ID               # subsection name
280 <int> 10                     # subsection type 10 = handle to owner entity, 3DSOLID/REGION
320 <str> 339                    # handle
2 <str> ASM_Data                 # subsection name
280 <int> 15                     # subsection type 15 = binary data
94 <int> 1088                    # size of data
310 <binary encoded data>        # data
310 <binary encoded data>        # data
...

0 <str> ENDSEC
"""
from typing import TYPE_CHECKING, Iterator, Iterable, List, Any, Union
import abc
from itertools import islice

from ezdxf.lldxf.tags import group_tags, Tags
from ezdxf.lldxf.const import DXFKeyError, DXFStructureError

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, Drawing


class AcDsEntity(abc.ABC):
    @abc.abstractmethod
    def export_dxf(self, tagwriter: "TagWriter"):
        ...

    @abc.abstractmethod
    def dxftype(self) -> str:
        ...


class AcDsDataSection:
    name = "ACDSDATA"

    def __init__(self, doc: "Drawing", entities: Iterable[Tags] = None):
        self.doc = doc
        self.entities: List[AcDsEntity] = []
        self.section_info = Tags()
        if entities is not None:
            self.load_tags(iter(entities))

    @property
    def is_valid(self):
        return len(self.section_info)

    def load_tags(self, entities: Iterator[Tags]) -> None:
        section_head = next(entities)
        if section_head[0] != (0, "SECTION") or section_head[1] != (
            2,
            "ACDSDATA",
        ):
            raise DXFStructureError(
                "Critical structure error in ACDSDATA section."
            )

        self.section_info = section_head
        for entity in entities:
            self.append(AcDsData(entity))  # tags have no subclasses

    def append(self, entity: "AcDsData") -> None:
        cls = ACDSDATA_TYPES.get(entity.dxftype(), AcDsData)
        data = cls(entity.tags)
        self.entities.append(data)

    def export_dxf(self, tagwriter: "TagWriter") -> None:
        if not self.is_valid:
            return
        tagwriter.write_tags(self.section_info)
        for entity in self.entities:
            entity.export_dxf(tagwriter)
        tagwriter.write_tag2(0, "ENDSEC")

    @property
    def acdsrecords(self) -> Iterable["AcDsRecord"]:
        return (
            entity for entity in self.entities if isinstance(entity, AcDsRecord)
        )

    def get_acis_data(self, handle: str) -> List[bytes]:
        for record in self.acdsrecords:
            try:
                section = record.get_section("AcDbDs::ID")
            except DXFKeyError:  # not present
                continue
            asm_handle = section.get_first_value(320, None)
            if asm_handle == handle:
                try:
                    asm_data = record.get_section("ASM_Data")
                except DXFKeyError:  # no data stored
                    break
                return [tag.value for tag in asm_data if tag.code == 310]
        return []


class AcDsData(AcDsEntity):
    def __init__(self, tags: Tags):
        self.tags = tags

    def export_dxf(self, tagwriter: "TagWriter"):
        tagwriter.write_tags(self.tags)

    def dxftype(self) -> str:
        return self.tags[0].value


class Section(Tags):
    @property
    def name(self) -> str:
        return self[0].value

    @property
    def type(self) -> str:
        return self[1].value

    @property
    def data(self) -> Tags:
        return Tags(self[2:])


class AcDsRecord(AcDsEntity):
    def __init__(self, tags: Tags):
        self._dxftype = tags[0]
        self.flags = tags[1]
        self.sections = [
            Section(group)
            for group in group_tags(islice(tags, 2, None), splitcode=2)
        ]

    def dxftype(self) -> str:
        return "ACDSRECORD"

    def has_section(self, name: str) -> bool:
        return self.get_section(name, default=None) is not None

    def get_section(self, name: str, default: Any = DXFKeyError) -> Section:
        for section in self.sections:
            if section.name == name:
                return section
        if default is DXFKeyError:
            raise DXFKeyError(name)
        else:
            return default

    def __len__(self):
        return len(self.sections)

    def __getitem__(self, item) -> Section:
        return self.sections[item]

    def _write_header(self, tagwriter: "TagWriter") -> None:
        tagwriter.write_tags(Tags([self._dxftype, self.flags]))

    def export_dxf(self, tagwriter: "TagWriter") -> None:
        self._write_header(tagwriter)
        for section in self.sections:
            tagwriter.write_tags(section)


ACDSDATA_TYPES = {
    "ACDSRECORD": AcDsRecord,
}



# Copyright (c) 2019-2021, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING
import logging
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.const import DXF12, SUBCLASS_MARKER, DXF2000, DXF2007
from ezdxf.lldxf.validator import is_valid_vport_name
from ezdxf.math import Vec2, NULLVEC, Z_AXIS
from ezdxf.entities.dxfentity import base_class, SubclassProcessor, DXFEntity
from ezdxf.entities.layer import acdb_symbol_table_record
from .factory import register_entity

logger = logging.getLogger("ezdxf")

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, DXFNamespace

__all__ = ["VPort"]

acdb_vport = DefSubclass(
    "AcDbViewportTableRecord",
    {
        "name": DXFAttr(2, validator=is_valid_vport_name),
        "flags": DXFAttr(70, default=0),
        "lower_left": DXFAttr(10, xtype=XType.point2d, default=Vec2(0, 0)),
        "upper_right": DXFAttr(11, xtype=XType.point2d, default=Vec2(1, 1)),
        "center": DXFAttr(12, xtype=XType.point2d, default=Vec2(344.2, 148.5)),
        "snap_base": DXFAttr(13, xtype=XType.point2d, default=Vec2(0, 0)),
        "snap_spacing": DXFAttr(
            14, xtype=XType.point2d, default=Vec2(0.5, 0.5)
        ),
        "grid_spacing": DXFAttr(
            15, xtype=XType.point2d, default=Vec2(0.5, 0.5)
        ),
        "direction": DXFAttr(
            16,
            xtype=XType.point3d,
            default=Z_AXIS,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
        "target": DXFAttr(17, xtype=XType.point3d, default=NULLVEC),
        # height: DXF reference error: listed as group code 45
        "height": DXFAttr(40, default=297),
        "aspect_ratio": DXFAttr(41, default=1.34),
        "focal_length": DXFAttr(42, default=50),
        "front_clipping": DXFAttr(43, default=0),
        "back_clipping": DXFAttr(44, default=0),
        "snap_rotation": DXFAttr(50, default=0),
        "view_twist": DXFAttr(51, default=0),
        "view_mode": DXFAttr(71, default=0),
        "circle_sides": DXFAttr(72, default=1000),
        "fast_zoom": DXFAttr(73, default=1),  # removed in R2007
        "ucs_icon": DXFAttr(74, default=3),
        "snap_on": DXFAttr(75, default=0),  # removed in R2007
        "grid_on": DXFAttr(76, default=0),  # removed in R2007
        "snap_style": DXFAttr(77, default=0),  # removed in R2007
        "snap_isopair": DXFAttr(78, default=0),  # removed in R2007
        # R2000: 331 or 441 (optional) - ignored by ezdxf
        # Soft or hard-pointer ID/handle to frozen layer objects;
        # repeats for each frozen layers
        # 70: Bit flags and perspective mode
        # CTB-File?
        "plot_style_sheet": DXFAttr(1, dxfversion=DXF2007),
        # Render mode:
        # 0 = 2D Optimized (classic 2D)
        # 1 = Wireframe
        # 2 = Hidden line
        # 3 = Flat shaded
        # 4 = Gouraud shaded
        # 5 = Flat shaded with wireframe
        # 6 = Gouraud shaded with wireframe
        # All rendering modes other than 2D Optimized engage the new 3D graphics
        # pipeline. These values directly correspond to the SHADEMODE command and
        # the AcDbAbstractViewTableRecord::RenderMode enum
        "render_mode": DXFAttr(
            281,
            default=0,
            dxfversion=DXF2000,
            validator=validator.is_in_integer_range(0, 7),
            fixer=RETURN_DEFAULT,
        ),
        # Value of UCSVP for this viewport. If set to 1, then viewport stores its
        # own UCS which will become the current UCS whenever the viewport is
        # activated. If set to 0, UCS will not change when this viewport is
        # activated
        "ucs_vp": DXFAttr(
            65,
            dxfversion=DXF2000,
            default=0,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        "ucs_origin": DXFAttr(110, xtype=XType.point3d, dxfversion=DXF2000),
        "ucs_xaxis": DXFAttr(
            111,
            xtype=XType.point3d,
            dxfversion=DXF2000,
            validator=validator.is_not_null_vector,
        ),
        "ucs_yaxis": DXFAttr(
            112,
            xtype=XType.point3d,
            dxfversion=DXF2000,
            validator=validator.is_not_null_vector,
        ),
        # Handle of AcDbUCSTableRecord if UCS is a named UCS. If not present,
        # then UCS is unnamed:
        "ucs_handle": DXFAttr(345, dxfversion=DXF2000),
        # Handle of AcDbUCSTableRecord of base UCS if UCS is orthographic (79 code
        # is non-zero). If not present and 79 code is non-zero, then base UCS is
        # taken to be WORLD:
        "base_ucs_handle": DXFAttr(346, dxfversion=DXF2000),
        # UCS ortho type:
        # 0 = UCS is not orthographic
        # 1 = Top
        # 2 = Bottom
        # 3 = Front
        # 4 = Back
        # 5 = Left
        # 6 = Right
        "ucs_ortho_type": DXFAttr(
            79,
            dxfversion=DXF2000,
            validator=validator.is_in_integer_range(0, 7),
        ),
        "elevation": DXFAttr(146, dxfversion=DXF2000, default=0),
        "unknown1": DXFAttr(60, dxfversion=DXF2000),
        "shade_plot_setting": DXFAttr(170, dxfversion=DXF2007),
        "major_grid_lines": DXFAttr(61, dxfversion=DXF2007),
        # Handle to background object
        "background_handle": DXFAttr(332, dxfversion=DXF2007, optional=True),
        # Handle to shade plot object
        "shade_plot_handle": DXFAttr(333, dxfversion=DXF2007, optional=True),
        # Handle to visual style object
        "visual_style_handle": DXFAttr(348, dxfversion=DXF2007, optional=True),
        "default_lighting_on": DXFAttr(
            292,
            dxfversion=DXF2007,
            validator=validator.is_integer_bool,
        ),
        # Default lighting type:
        # 0 = One distant light
        # 1 = Two distant lights
        "default_lighting_type": DXFAttr(
            282,
            dxfversion=DXF2007,
            validator=validator.is_integer_bool,
        ),
        "brightness": DXFAttr(141, dxfversion=DXF2000),
        "contrast": DXFAttr(142, dxfversion=DXF2000),
        "ambient_color_aci": DXFAttr(63, dxfversion=DXF2000, optional=True),
        "ambient_true_color": DXFAttr(421, dxfversion=DXF2000, optional=True),
        "ambient_color_name": DXFAttr(431, dxfversion=DXF2000, optional=True),
        # Hard-pointer handle to sun object:
        "sun_handle": DXFAttr(361, dxfversion=DXF2007, optional=True),
    },
)
acdb_vport_group_codes = group_code_mapping(acdb_vport)


@register_entity
class VPort(DXFEntity):
    """DXF VIEW entity"""

    DXFTYPE = "VPORT"
    DXFATTRIBS = DXFAttributes(base_class, acdb_symbol_table_record, acdb_vport)

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_vport_group_codes, subclass=2
            )
        return dxf

    def export_entity(self, tagwriter: "TagWriter") -> None:
        super().export_entity(tagwriter)
        dxfversion = tagwriter.dxfversion
        if dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_symbol_table_record.name)
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_vport.name)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "name",
                "flags",
                "lower_left",
                "upper_right",
                "center",
                "snap_base",
                "snap_spacing",
                "grid_spacing",
                "direction",
                "target",
                "height",
                "aspect_ratio",
                "focal_length",
                "front_clipping",
                "back_clipping",
                "snap_rotation",
                "view_twist",
                "view_mode",
                "circle_sides",
                "fast_zoom",
                "ucs_icon",
                "snap_on",
                "grid_on",
                "snap_style",
                "snap_isopair",
                "plot_style_sheet",
                "render_mode",
                "ucs_vp",
                "ucs_origin",
                "ucs_xaxis",
                "ucs_yaxis",
                "ucs_handle",
                "base_ucs_handle",
                "ucs_ortho_type",
                "elevation",
                "unknown1",
                "shade_plot_setting",
                "major_grid_lines",
                "background_handle",
                "shade_plot_handle",
                "visual_style_handle",
                "default_lighting_on",
                "default_lighting_type",
                "brightness",
                "contrast",
                "ambient_color_aci",
                "ambient_true_color",
                "ambient_color_name",
            ],
        )



# Copyright (c) 2019-2021, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Optional
import logging
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.const import (
    DXF12,
    SUBCLASS_MARKER,
    DXF2007,
    DXFInternalEzdxfError,
)
from ezdxf.entities.dxfentity import base_class, SubclassProcessor, DXFEntity
from ezdxf.entities.layer import acdb_symbol_table_record
from .factory import register_entity

logger = logging.getLogger("ezdxf")

if TYPE_CHECKING:
    from ezdxf.eztypes import (
        TagWriter,
        DXFNamespace,
        DXFGraphic,
        EntitySpace,
        BlockLayout,
        Block,
        EndBlk
    )

__all__ = ["BlockRecord"]

acdb_blockrec = DefSubclass(
    "AcDbBlockTableRecord",
    {
        "name": DXFAttr(2, validator=validator.is_valid_block_name),
        # handle to associated DXF LAYOUT object
        "layout": DXFAttr(340, default="0"),
        # 0 = can not explode; 1 = can explode
        "explode": DXFAttr(
            280,
            default=1,
            dxfversion=DXF2007,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        # 0 = scale non uniformly; 1 = scale uniformly
        "scale": DXFAttr(
            281,
            default=0,
            dxfversion=DXF2007,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        # see ezdxf/units.py
        "units": DXFAttr(
            70,
            default=0,
            dxfversion=DXF2007,
            validator=validator.is_in_integer_range(0, 25),
            fixer=RETURN_DEFAULT,
        ),
        # 310: Binary data for bitmap preview (optional) - removed (ignored) by ezdxf
    },
)
acdb_blockrec_group_codes = group_code_mapping(acdb_blockrec)

# optional XDATA for all DXF versions
# 1000: "ACAD"
# 1001: "DesignCenter Data" (optional)
# 1002: "{"
# 1070: Autodesk Design Center version number
# 1070: Insert units: like 'units'
# 1002: "}"


@register_entity
class BlockRecord(DXFEntity):
    """DXF BLOCK_RECORD table entity

    BLOCK_RECORD is the hard owner of all entities in BLOCK definitions, this
    means owner tag of entities is handle of BLOCK_RECORD.

    """

    DXFTYPE = "BLOCK_RECORD"
    DXFATTRIBS = DXFAttributes(
        base_class, acdb_symbol_table_record, acdb_blockrec
    )

    def __init__(self):
        from ezdxf.entitydb import EntitySpace

        super().__init__()
        # Store entities in the block_record instead of BlockLayout and Layout,
        # because BLOCK_RECORD is also the hard owner of all the entities.
        self.entity_space = EntitySpace()
        self.block: Optional["Block"] = None
        self.endblk: Optional["EndBlk"] = None
        # stores also the block layout structure
        self.block_layout: Optional[BlockLayout] = None

    def set_block(self, block: "Block", endblk: "EndBlk"):
        self.block = block
        self.endblk = endblk
        self.block.dxf.owner = self.dxf.handle
        self.endblk.dxf.owner = self.dxf.handle

    def set_entity_space(self, entity_space: "EntitySpace") -> None:
        self.entity_space = entity_space

    def rename(self, name: str) -> None:
        assert self.block is not None
        self.dxf.name = name
        self.block.dxf.name = name

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(dxf, acdb_blockrec_group_codes, 2)
        return dxf

    def export_entity(self, tagwriter: "TagWriter") -> None:
        super().export_entity(tagwriter)
        if tagwriter.dxfversion == DXF12:
            raise DXFInternalEzdxfError("Exporting BLOCK_RECORDS for DXF R12.")
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_symbol_table_record.name)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_blockrec.name)

        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "name",
                "layout",
                "units",
                "explode",
                "scale",
            ],
        )

    def export_block_definition(self, tagwriter: "TagWriter") -> None:
        """Exports BLOCK, than all DXF entities and at last the ENDBLK entity,
        except for *Model_space and *Paper_Pacer, their entities are stored
        in the entities section.

        """
        assert self.block is not None
        assert self.endblk is not None
        if self.block_layout is not None:
            self.block_layout.update_block_flags()
        self.block.export_dxf(tagwriter)
        if not (self.is_modelspace or self.is_active_paperspace):
            self.entity_space.export_dxf(tagwriter)
        self.endblk.export_dxf(tagwriter)

    def destroy(self):
        """Destroy associated data:

            - BLOCK
            - ENDBLK
            - all entities stored in this block definition

        Does not destroy the linked LAYOUT entity, this is the domain of the
        :class:`Layouts` object, which also should initiate the destruction of
        'this' BLOCK_RECORD.

        """
        if not self.is_alive:
            return

        self.block.destroy()
        self.endblk.destroy()
        for entity in self.entity_space:
            entity.destroy()

        # remove attributes to find invalid access after death
        del self.block
        del self.endblk
        del self.block_layout
        super().destroy()

    @property
    def is_active_paperspace(self) -> bool:
        """``True`` if is "active" paperspace layout."""
        return self.dxf.name.lower() == "*paper_space"

    @property
    def is_any_paperspace(self) -> bool:
        """``True`` if is any kind of paperspace layout."""
        return self.dxf.name.lower().startswith("*paper_space")

    @property
    def is_modelspace(self) -> bool:
        """``True`` if is the modelspace layout."""
        return self.dxf.name.lower() == "*model_space"

    @property
    def is_any_layout(self) -> bool:
        """``True`` if is any kind of modelspace or paperspace layout."""
        return self.is_modelspace or self.is_any_paperspace

    @property
    def is_block_layout(self) -> bool:
        """``True`` if not any kind of modelspace or paperspace layout, just a
        regular block definition.
        """
        return not self.is_any_layout

    def add_entity(self, entity: "DXFGraphic") -> None:
        """Add an existing DXF entity to BLOCK_RECORD.

        Args:
            entity: :class:`DXFGraphic`

        """
        # assign layout
        try:
            entity.set_owner(
                self.dxf.handle, paperspace=int(self.is_any_paperspace)
            )
        except AttributeError:
            logger.debug(
                f"Unexpected DXF entity {str(entity)} in {str(self.block)}"
            )
        # Add unexpected entities also to the entity space - auditor should fix
        # errors!
        self.entity_space.add(entity)

    def unlink_entity(self, entity: "DXFGraphic") -> None:
        """Unlink `entity` from BLOCK_RECORD.

        Removes `entity` just from  entity space but not from the drawing
        database.

        Args:
            entity: :class:`DXFGraphic`

        """
        if entity.is_alive:
            self.entity_space.remove(entity)
            try:
                entity.set_owner(None)
            except AttributeError:
                pass  # unsupported entities as DXFTagStorage

    def delete_entity(self, entity: "DXFGraphic") -> None:
        """Delete `entity` from BLOCK_RECORD entity space and drawing database.

        Args:
            entity: :class:`DXFGraphic`

        """
        self.unlink_entity(entity)  # 1. unlink from entity space
        entity.destroy()



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
#
# Pure Python implementation of the B-spline basis function.

from typing import List, Iterable, Sequence, Tuple
import math
import bisect

# The pure Python implementation can't import from ._ctypes or ezdxf.math!
from ._vector import Vec3, NULLVEC
from .linalg import binomial_coefficient

__all__ = ["Basis", "Evaluator"]


class Basis:
    """Immutable Basis function class."""

    __slots__ = ("_knots", "_weights", "_order", "_count")

    def __init__(
        self,
        knots: Iterable[float],
        order: int,
        count: int,
        weights: Sequence[float] = None,
    ):
        self._knots = tuple(knots)
        self._weights = tuple(weights or [])
        self._order: int = int(order)
        self._count: int = int(count)

        # validation checks:
        len_weights = len(self._weights)
        if len_weights != 0 and len_weights != self._count:
            raise ValueError("invalid weight count")
        if len(self._knots) != self._order + self._count:
            raise ValueError("invalid knot count")

    @property
    def max_t(self) -> float:
        return self._knots[-1]

    @property
    def order(self) -> int:
        return self._order

    @property
    def degree(self) -> int:
        return self._order - 1

    @property
    def knots(self) -> Tuple[float, ...]:
        return self._knots

    @property
    def weights(self) -> Tuple[float, ...]:
        return self._weights

    @property
    def is_rational(self) -> bool:
        """Returns ``True`` if curve is a rational B-spline. (has weights)"""
        return bool(self._weights)

    def basis_vector(self, t: float) -> List[float]:
        """Returns the expanded basis vector."""
        span = self.find_span(t)
        p = self._order - 1
        front = span - p
        back = self._count - span - 1
        basis = self.basis_funcs(span, t)
        return ([0.0] * front) + basis + ([0.0] * back)

    def find_span(self, u: float) -> int:
        """Determine the knot span index."""
        # Linear search is more reliable than binary search of the Algorithm A2.1
        # from The NURBS Book by Piegl & Tiller.
        knots = self._knots
        count = self._count  # text book: n+1
        if u >= knots[count]:  # special case
            return count - 1  # n
        p = self._order - 1
        # common clamped spline:
        if knots[p] == 0.0:  # use binary search
            # This is fast and works most of the time,
            # but Test 621 : test_weired_closed_spline()
            # goes into an infinity loop, because of
            # a weird knot configuration.
            return bisect.bisect_right(knots, u, p, count) - 1
        else:  # use linear search
            span = 0
            while knots[span] <= u and span < count:
                span += 1
            return span - 1

    def basis_funcs(self, span: int, u: float) -> List[float]:
        # Source: The NURBS Book: Algorithm A2.2
        order = self._order
        knots = self._knots
        N = [0.0] * order
        left = list(N)
        right = list(N)
        N[0] = 1.0
        for j in range(1, order):
            left[j] = u - knots[max(0, span + 1 - j)]
            right[j] = knots[span + j] - u
            saved = 0.0
            for r in range(j):
                temp = N[r] / (right[r + 1] + left[j - r])
                N[r] = saved + right[r + 1] * temp
                saved = left[j - r] * temp
            N[j] = saved
        if self.is_rational:
            return self.span_weighting(N, span)
        else:
            return N

    def span_weighting(self, nbasis: List[float], span: int) -> List[float]:
        size = len(nbasis)
        weights = self._weights[span - self._order + 1 : span + 1]
        products = [nb * w for nb, w in zip(nbasis, weights)]
        s = sum(products)
        return [0.0] * size if s == 0.0 else [p / s for p in products]

    def basis_funcs_derivatives(self, span: int, u: float, n: int = 1):
        # Source: The NURBS Book: Algorithm A2.3
        order = self._order
        p = order - 1
        n = min(n, p)

        knots = self._knots
        left = [1.0] * order
        right = [1.0] * order
        ndu = [[1.0] * order for _ in range(order)]

        for j in range(1, order):
            left[j] = u - knots[max(0, span + 1 - j)]
            right[j] = knots[span + j] - u
            saved = 0.0
            for r in range(j):
                # lower triangle
                ndu[j][r] = right[r + 1] + left[j - r]
                temp = ndu[r][j - 1] / ndu[j][r]
                # upper triangle
                ndu[r][j] = saved + (right[r + 1] * temp)
                saved = left[j - r] * temp
            ndu[j][j] = saved

        # load the basis_vector functions
        derivatives = [[0.0] * order for _ in range(order)]
        for j in range(order):
            derivatives[0][j] = ndu[j][p]

        # loop over function index
        a = [[1.0] * order, [1.0] * order]
        for r in range(order):
            s1 = 0
            s2 = 1
            # alternate rows in array a
            a[0][0] = 1.0

            # loop to compute kth derivative
            for k in range(1, n + 1):
                d = 0.0
                rk = r - k
                pk = p - k
                if r >= k:
                    a[s2][0] = a[s1][0] / ndu[pk + 1][rk]
                    d = a[s2][0] * ndu[rk][pk]
                if rk >= -1:
                    j1 = 1
                else:
                    j1 = -rk
                if (r - 1) <= pk:
                    j2 = k - 1
                else:
                    j2 = p - r
                for j in range(j1, j2 + 1):
                    a[s2][j] = (a[s1][j] - a[s1][j - 1]) / ndu[pk + 1][rk + j]
                    d += a[s2][j] * ndu[rk + j][pk]
                if r <= pk:
                    a[s2][k] = -a[s1][k - 1] / ndu[pk + 1][r]
                    d += a[s2][k] * ndu[r][pk]
                derivatives[k][r] = d

                # Switch rows
                s1, s2 = s2, s1

        # Multiply through by the the correct factors
        r = float(p)  # type: ignore
        for k in range(1, n + 1):
            for j in range(order):
                derivatives[k][j] *= r
            r *= p - k
        return derivatives[: n + 1]


class Evaluator:
    """B-spline curve point and curve derivative evaluator."""

    __slots__ = ["_basis", "_control_points"]

    def __init__(self, basis: Basis, control_points: Sequence[Vec3]):
        self._basis = basis
        self._control_points = control_points

    def point(self, u: float) -> Vec3:
        # Source: The NURBS Book: Algorithm A3.1
        basis = self._basis
        control_points = self._control_points
        if math.isclose(u, basis.max_t):
            u = basis.max_t

        p = basis.degree
        span = basis.find_span(u)
        N = basis.basis_funcs(span, u)
        return Vec3.sum(
            N[i] * control_points[span - p + i] for i in range(p + 1)
        )

    def points(self, t: Iterable[float]) -> Iterable[Vec3]:
        for u in t:
            yield self.point(u)

    def derivative(self, u: float, n: int = 1) -> List[Vec3]:
        """Return point and derivatives up to n <= degree for parameter u."""
        # Source: The NURBS Book: Algorithm A3.2
        basis = self._basis
        control_points = self._control_points
        if math.isclose(u, basis.max_t):
            u = basis.max_t

        p = basis.degree
        span = basis.find_span(u)
        basis_funcs_ders = basis.basis_funcs_derivatives(span, u, n)
        if basis.is_rational:
            # Homogeneous point representation required:
            # (x*w, y*w, z*w, w)
            CKw: List[Vec3] = []
            wders: List[float] = []
            weights = basis.weights
            for k in range(n + 1):
                v = NULLVEC
                wder = 0.0
                for j in range(p + 1):
                    index = span - p + j
                    bas_func_weight = basis_funcs_ders[k][j] * weights[index]
                    # control_point * weight * bas_func_der = (x*w, y*w, z*w) * bas_func_der
                    v += control_points[index] * bas_func_weight
                    wder += bas_func_weight
                CKw.append(v)
                wders.append(wder)

            # Source: The NURBS Book: Algorithm A4.2
            CK: List[Vec3] = []
            for k in range(n + 1):
                v = CKw[k]
                for i in range(1, k + 1):
                    v -= binomial_coefficient(k, i) * wders[i] * CK[k - i]
                CK.append(v / wders[0])
        else:
            CK = [
                Vec3.sum(
                    basis_funcs_ders[k][j] * control_points[span - p + j]
                    for j in range(p + 1)
                )
                for k in range(n + 1)
            ]
        return CK

    def derivatives(
        self, t: Iterable[float], n: int = 1
    ) -> Iterable[List[Vec3]]:
        for u in t:
            yield self.derivative(u, n)



# Copyright (c) 2018-2021 Manfred Moitzi
# License: MIT License
"""
DXF R12 Splines
===============

DXF R12 supports 2d B-splines, but Autodesk do not document the usage in the
DXF Reference. The base entity for splines in DXF R12 is the POLYLINE entity.

Transformed Into 3D Space
-------------------------

The spline itself is always in a plane, but as any 2D entity, the spline can be
transformed into the 3D object by elevation, extrusion and thickness/width.

Open Quadratic Spline with Fit Vertices
-------------------------------------

Example: 2D_SPLINE_QUADRATIC.dxf
expected knot vector: open uniform
degree: 2
order: 3

POLYLINE:
flags (70): 4 = SPLINE_FIT_VERTICES_ADDED
smooth type (75): 5 = QUADRATIC_BSPLINE

Sequence of VERTEX
flags (70): SPLINE_VERTEX_CREATED = 8  # Spline vertex created by spline-fitting

This vertices are the curve vertices of the spline (fitted).

Frame control vertices appear after the curve vertices.

Sequence of VERTEX
flags (70): SPLINE_FRAME_CONTROL_POINT = 16

No control point at the starting point, but a control point at the end point,
last control point == last fit vertex

Closed Quadratic Spline with Fit Vertices
-----------------------------------------

Example: 2D_SPLINE_QUADRATIC_CLOSED.dxf
expected knot vector: closed uniform
degree: 2
order: 3

POLYLINE:
flags (70): 5 = CLOSED | SPLINE_FIT_VERTICES_ADDED
smooth type (75): 5 = QUADRATIC_BSPLINE

Sequence of VERTEX
flags (70): SPLINE_VERTEX_CREATED = 8  # Spline vertex created by spline-fitting

Frame control vertices appear after the curve vertices.

Sequence of VERTEX
flags (70): SPLINE_FRAME_CONTROL_POINT = 16


Open Cubic Spline with Fit Vertices
-----------------------------------

Example: 2D_SPLINE_CUBIC.dxf
expected knot vector: open uniform
degree: 3
order: 4

POLYLINE:
flags (70): 4 = SPLINE_FIT_VERTICES_ADDED
smooth type (75): 6 = CUBIC_BSPLINE

Sequence of VERTEX
flags (70): SPLINE_VERTEX_CREATED = 8  # Spline vertex created by spline-fitting

This vertices are the curve vertices of the spline (fitted).

Frame control vertices appear after the curve vertices.

Sequence of VERTEX
flags (70): SPLINE_FRAME_CONTROL_POINT = 16

No control point at the starting point, but a control point at the end point,
last control point == last fit vertex

Closed Curve With Extra Vertices Created
----------------------------------------

Example: 2D_FIT_CURVE_CLOSED.dxf

POLYLINE:
flags (70): 3 = CLOSED | CURVE_FIT_VERTICES_ADDED

Vertices with bulge values:

flags (70): 1 = EXTRA_VERTEX_CREATED
Vertex 70=0, Vertex 70=1, Vertex 70=0, Vertex 70=1

"""
from typing import TYPE_CHECKING, Iterable, List
from ezdxf.lldxf import const
from ezdxf.math import BSpline, closed_uniform_bspline, Vec3

if TYPE_CHECKING:
    from ezdxf.eztypes import Vertex, BaseLayout, Polyline
    from ezdxf.math.ucs import UCS


class R12Spline:
    """DXF R12 supports 2D B-splines, but Autodesk do not document the usage
    in the DXF Reference. The base entity for splines in DXF R12 is the POLYLINE
    entity. The spline itself is always in a plane, but as any 2D entity, the
    spline can be transformed into the 3D object by elevation and extrusion
    (:ref:`OCS`, :ref:`UCS`).

    This way it was possible to store the spline parameters in the DXF R12 file,
    to allow CAD applications to modify the spline parameters and rerender the
    B-spline afterward again as polyline approximation. Therefore the result is
    not better than an approximation by the :class:`Spline` class, it is also
    just a POLYLINE entity, but maybe someone need exact this tool in the
    future.

    """

    def __init__(
        self,
        control_points: Iterable["Vertex"],
        degree: int = 2,
        closed: bool = True,
    ):
        """
        Args:
            control_points: B-spline control frame vertices
            degree: degree of B-spline, only 2 and 3 is supported
            closed: ``True`` for closed curve

        """
        self.control_points = Vec3.list(control_points)
        self.degree = degree
        self.closed = closed

    def approximate(
        self, segments: int = 40, ucs: "UCS" = None
    ) -> List["Vertex"]:
        """Approximate the B-spline by a polyline with `segments` line segments.
        If `ucs` is not ``None``, ucs defines an :class:`~ezdxf.math.UCS`, to
        transformed the curve into :ref:`OCS`. The control points are placed
        xy-plane of the UCS, don't use z-axis coordinates, if so make sure all
        control points are in a plane parallel to the OCS base plane
        (UCS xy-plane), else the result is unpredictable and depends on the CAD
        application used to open the DXF file - it may crash.

        Args:
            segments: count of line segments for approximation, vertex count is
                `segments` + 1
            ucs: :class:`~ezdxf.math.UCS` definition, control points in ucs
                coordinates

        Returns:
            list of vertices in :class:`~ezdxf.math.OCS` as
            :class:`~ezdxf.math.Vec3` objects

        """
        if self.closed:
            spline = closed_uniform_bspline(
                self.control_points, order=self.degree + 1
            )
        else:
            spline = BSpline(self.control_points, order=self.degree + 1)
        vertices = spline.approximate(segments)
        if ucs is not None:
            vertices = (ucs.to_ocs(vertex) for vertex in vertices)
        return list(vertices)

    def render(
        self,
        layout: "BaseLayout",
        segments: int = 40,
        ucs: "UCS" = None,
        dxfattribs: dict = None,
    ) -> "Polyline":
        """Renders the B-spline into `layout` as 2D :class:`~ezdxf.entities.Polyline`
        entity. Use an :class:`~ezdxf.math.UCS` to place the 2D spline in the
        3D space, see :meth:`approximate` for more information.

        Args:
            layout: :class:`~ezdxf.layouts.BaseLayout` object
            segments: count of line segments for approximation, vertex count is
                `segments` + 1
            ucs: :class:`~ezdxf.math.UCS` definition, control points in ucs
                coordinates.
            dxfattribs: DXF attributes for :class:`~ezdxf.entities.Polyline`

        """
        polyline = layout.add_polyline2d(points=[], dxfattribs=dxfattribs)
        flags = polyline.SPLINE_FIT_VERTICES_ADDED
        if self.closed:
            flags |= polyline.CLOSED
        polyline.dxf.flags = flags

        if self.degree == 2:
            smooth_type = polyline.QUADRATIC_BSPLINE
        elif self.degree == 3:
            smooth_type = polyline.CUBIC_BSPLINE
        else:
            raise ValueError("invalid degree of spline")
        polyline.dxf.smooth_type = smooth_type

        # set OCS extrusion vector
        if ucs is not None:
            polyline.dxf.extrusion = ucs.uz

        # add fit points in OCS
        polyline.append_vertices(
            self.approximate(segments, ucs),
            dxfattribs={
                "layer": polyline.dxf.layer,
                "flags": const.VTX_SPLINE_VERTEX_CREATED,
            },
        )

        # add control frame points in OCS
        control_points = self.control_points
        if ucs is not None:
            control_points = list(ucs.points_to_ocs(control_points))
            polyline.dxf.elevation = (0, 0, control_points[0].z)
        polyline.append_vertices(
            control_points,
            dxfattribs={
                "layer": polyline.dxf.layer,
                "flags": const.VTX_SPLINE_FRAME_CONTROL_POINT,
            },
        )
        return polyline



# Copyright (c) 2016-2021, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, List, Tuple, Sequence, Union, cast
from ezdxf._options import options
from ezdxf.render.arrows import ARROWS
from ezdxf.lldxf.const import DXF12
import logging

if TYPE_CHECKING:
    from ezdxf.eztypes import Drawing, DimStyle

logger = logging.getLogger("ezdxf")
LTypeDef = Tuple[str, str, Sequence[float]]


def setup_drawing(doc: "Drawing", topics: Union[str, bool, Sequence] = "all"):
    """
    Setup default linetypes, text styles or dimension styles.

    Args:
        doc: DXF document
        topics: 'all' or True to setup everything
                Tuple of strings to specify setup:
                    - 'linetypes': setup linetypes
                    - 'styles': setup text styles
                    - 'dimstyles[:all|metric|us]': setup dimension styles (us not implemented)
                    - 'visualstyles': setup 25 standard visual styles

    """
    if not topics:  # topics is None, False or ''
        return

    def get_token(name: str) -> List[str]:
        for t in topics:  # type: ignore
            token = t.split(":")
            if token[0] == name:
                return token
        return []

    if topics in ("all", True):
        setup_all = True
        topics = []
    else:
        setup_all = False
        topics = list(t.lower() for t in topics)  # type: ignore

    if setup_all or "linetypes" in topics:
        setup_linetypes(doc)

    if setup_all or "styles" in topics:
        setup_styles(doc)

    if setup_all or "visualstyles" in topics:
        setup_visual_styles(doc)

    dimstyles = get_token("dimstyles")
    if setup_all or len(dimstyles):
        if len(dimstyles) == 2:
            domain = dimstyles[1]
        else:
            domain = "all"
        setup_dimstyles(doc, domain=domain)


def setup_linetypes(doc: "Drawing") -> None:
    measurement = 1
    if doc:
        measurement = doc.header.get("$MEASUREMENT", measurement)
    factor = ISO_LTYPE_FACTOR if measurement else 1.0
    for name, desc, pattern in linetypes(scale=factor):
        if name in doc.linetypes:
            continue
        doc.linetypes.new(
            name,
            dxfattribs={
                "description": desc,
                "pattern": pattern,
            },
        )


def setup_styles(doc: "Drawing") -> None:
    doc.header["$TEXTSTYLE"] = "OpenSans"
    for name, font in styles():
        if name in doc.styles:
            continue
        doc.styles.new(
            name,
            dxfattribs={
                "font": font,
            },
        )


def setup_dimstyles(doc: "Drawing", domain: str = "all") -> None:
    setup_styles(doc)
    ezdxf_dimstyle = setup_dimstyle(
        doc,
        name="EZDXF",
        fmt="EZ_M_100_H25_CM",
        style=options.default_dimension_text_style,
        blk=ARROWS.architectural_tick,
    )
    ezdxf_dimstyle.dxf.dimasz *= 0.7  # smaller arch ticks
    doc.header["$DIMSTYLE"] = "EZDXF"
    ezdxf_dimstyle.copy_to_header(doc)

    if domain in ("metric", "all"):
        setup_dimstyle(
            doc,
            fmt="EZ_M_100_H25_CM",
            style=options.default_dimension_text_style,
        )
        setup_dimstyle(
            doc,
            fmt="EZ_M_50_H25_CM",
            style=options.default_dimension_text_style,
        )
        setup_dimstyle(
            doc,
            fmt="EZ_M_25_H25_CM",
            style=options.default_dimension_text_style,
        )
        setup_dimstyle(
            doc,
            fmt="EZ_M_20_H25_CM",
            style=options.default_dimension_text_style,
        )
        setup_dimstyle(
            doc,
            fmt="EZ_M_10_H25_CM",
            style=options.default_dimension_text_style,
        )
        setup_dimstyle(
            doc, fmt="EZ_M_5_H25_CM", style=options.default_dimension_text_style
        )
        setup_dimstyle(
            doc, fmt="EZ_M_1_H25_CM", style=options.default_dimension_text_style
        )
    elif domain in ("us", "all"):
        pass
    if domain in ("radius", "all"):
        ez_radius = cast(
            "DimStyle", doc.dimstyles.duplicate_entry("EZDXF", "EZ_RADIUS")
        )
        ez_radius.set_arrows(blk=ARROWS.closed_blank)
        ez_radius.dxf.dimasz = 0.25  # set arrow size
        ez_radius.dxf.dimtofl = 0  # force dimension line if text outside
        ez_radius.dxf.dimcen = (
            0  # size of center mark, 0=disable, >0=draw mark, <0=draw lines
        )
        # dimtmove: use leader, is the best setting for text outside to preserves
        # appearance of DIMENSION entity,if editing DIMENSION afterwards in
        # BricsCAD (AutoCAD)
        ez_radius.dxf.dimtmove = 1

        ez_radius_inside = doc.dimstyles.duplicate_entry(
            "EZ_RADIUS", "EZ_RADIUS_INSIDE"
        )
        # dimtmove: keep dim line with text, is the best setting for text inside
        # to preserves appearance of DIMENSION entity, if editing DIMENSION
        # afterwards in BricsCAD (AutoCAD)
        ez_radius_inside.dxf.dimtmove = 0
        ez_radius_inside.dxf.dimtix = 1  # force text inside
        ez_radius_inside.dxf.dimatfit = (
            0  # required by BricsCAD (AutoCAD) to force text inside
        )
        ez_radius_inside.dxf.dimtad = 0  # center text vertical


class DimStyleFmt:
    DIMASZ = 2.5  # in mm in paper space
    DIMTSZ = 1.25  # x2 in mm in paper space
    UNIT_FACTOR = {
        "m": 1,  # 1 drawing unit == 1 meter
        "dm": 10,  # 1 drawing unit == 1 decimeter
        "cm": 100,  # 1 drawing unit == 1 centimeter
        "mm": 1000,  # 1 drawing unit == 1 millimeter
    }

    def __init__(self, fmt: str):
        tokens = fmt.lower().split("_")
        self.name = fmt
        self.drawing_unit = tokens[1]  # EZ_<M>_100_H25_CM
        self.scale = float(tokens[2])  # EZ_M_<100>_H25_CM
        self.height = float(tokens[3][1:]) / 10.0  # EZ_M_100_H<25>_CM  # in mm
        self.measurement_unit = tokens[4]  # EZ_M_100_H25_<CM>

    @property
    def unit_factor(self):
        return self.UNIT_FACTOR[self.drawing_unit]

    @property
    def measurement_factor(self):
        return self.UNIT_FACTOR[self.measurement_unit]

    @property
    def text_factor(self):
        return self.unit_factor / self.UNIT_FACTOR["mm"] * self.scale

    @property
    def dimlfac(self):
        return self.measurement_factor / self.unit_factor

    @property
    def dimasz(self):
        return self.DIMASZ * self.text_factor

    @property
    def dimtsz(self):
        return self.DIMTSZ * self.text_factor

    @property
    def dimtxt(self):
        return self.height * self.text_factor

    @property
    def dimexe(self):
        return self.dimtxt * 1.5

    @property
    def dimexo(self):
        return self.dimtxt / 2

    @property
    def dimdle(self):
        return 0.25 * self.unit_factor


def setup_dimstyle(
    doc: "Drawing", fmt: str, style: str = None, blk: str = None, name: str = ""
) -> "DimStyle":
    """Easy DimStyle setup, the `fmt` string defines four essential dimension
    parameters separated by the `_` character. Tested and works with the metric
    system, I don't touch the 'english unit' system.

    Example: `fmt` = 'EZ_M_100_H25_CM'

        1. '<EZ>_M_100_H25_CM': arbitrary prefix
        2. 'EZ_<M>_100_H25_CM': defines the drawing unit, valid values are 'M', 'DM', 'CM', 'MM'
        3. 'EZ_M_<100>_H25_CM': defines the scale of the drawing, '100' is for 1:100
        4. 'EZ_M_100_<H25>_CM': defines the text height in mm in paper space times 10, 'H25' is 2.5mm
        5. 'EZ_M_100_H25_<CM>': defines the units for the measurement text, valid values are 'M', 'DM', 'CM', 'MM'

    Args:
        doc: DXF drawing
        fmt: format string
        style: text style for measurement
        blk: block name for arrow None for oblique stroke
        name: dimension style name, if name is '', `fmt` string is used as name

    """
    style = style or options.default_dimension_text_style
    dim_style_fmt = DimStyleFmt(fmt)
    name = name or dim_style_fmt.name
    if doc.dimstyles.has_entry(name):
        logging.debug('DimStyle "{}" already exists.'.format(name))
        return cast("DimStyle", doc.dimstyles.get(name))

    dimstyle = cast("DimStyle", doc.dimstyles.new(name))
    dimstyle.dxf.dimtxt = dim_style_fmt.dimtxt
    dimstyle.dxf.dimlfac = (
        dim_style_fmt.dimlfac
    )  # factor for measurement; dwg in m : measurement in cm -> dimlfac=100
    dimstyle.dxf.dimgap = (
        dim_style_fmt.dimtxt * 0.4
    )  # gap between text and dimension line
    dimstyle.dxf.dimtad = 1  # text above dimline
    dimstyle.dxf.dimexe = dim_style_fmt.dimexe
    dimstyle.dxf.dimexo = dim_style_fmt.dimexo
    dimstyle.dxf.dimdle = 0  # dimension extension beyond extension lines
    dimstyle.dxf.dimtix = 0  # Draws dimension text between the extension lines even if it would ordinarily be placed outside those lines
    dimstyle.dxf.dimtih = 0  # Aligns text inside extension lines with dimension line; 1 = Draws text horizontally
    dimstyle.dxf.dimtoh = 0  # Aligns text outside of extension lines with dimension line; 1 = Draws text horizontally
    dimstyle.dxf.dimzin = 8  # Suppresses trailing zeros in decimal dimensions
    dimstyle.dxf.dimsah = 0
    if blk is None:  # oblique stroke
        dimstyle.dxf.dimtsz = dim_style_fmt.dimtsz  # tick size
        dimstyle.dxf.dimasz = dim_style_fmt.dimasz  # arrow size
    else:  # arrow or block
        dimstyle.set_arrows(blk=blk)
        dimstyle.dxf.dimasz = dim_style_fmt.dimasz
    if doc.dxfversion > DXF12:
        # set text style
        dimstyle.dxf.dimtmove = 2  # move freely without leader
        dimstyle.dxf.dimtxsty = style

        # user location override, controls both the text position and the
        # dimension line location, same as DXF12
        dimstyle.dxf.dimupt = 1
        dimstyle.dxf.dimdsep = ord(".")
        dimstyle.dxf.dimdec = 2  # show just 2 decimals
    return dimstyle


ISO_LTYPE_FACTOR = 2.54
# DXF linetype definition for $MEASUREMENT=0 (imperial)
# name, description, elements:
# elements = [total_pattern_length, elem1, elem2, ...]
# total_pattern_length = sum(abs(elem))
# elem > 0 is line, < 0 is gap, 0.0 = dot;
ANSI_LINE_TYPES = [
    ("CONTINUOUS", "Solid", [0.0]),
    (
        "CENTER",
        "Center ____ _ ____ _ ____ _ ____ _ ____ _ ____",
        [2.0, 1.25, -0.25, 0.25, -0.25],
    ),
    (
        "CENTERX2",
        "Center (2x) ________  __  ________  __  ________",
        [3.5, 2.5, -0.25, 0.5, -0.25],
    ),
    (
        "CENTER2",
        "Center (.5x) ____ _ ____ _ ____ _ ____ _ ____",
        [1.0, 0.625, -0.125, 0.125, -0.125],
    ),
    (
        "DASHED",
        "Dashed __ __ __ __ __ __ __ __ __ __ __ __ __ _",
        [0.6, 0.5, -0.1],
    ),
    (
        "DASHEDX2",
        "Dashed (2x) ____  ____  ____  ____  ____  ____",
        [1.2, 1.0, -0.2],
    ),
    ("DASHED2", "Dashed (.5x) _ _ _ _ _ _ _ _ _ _ _ _ _ _", [0.3, 0.25, -0.05]),
    (
        "PHANTOM",
        "Phantom ______  __  __  ______  __  __  ______",
        [2.5, 1.25, -0.25, 0.25, -0.25, 0.25, -0.25],
    ),
    (
        "PHANTOMX2",
        "Phantom (2x)____________    ____    ____    ____________",
        [4.25, 2.5, -0.25, 0.5, -0.25, 0.5, -0.25],
    ),
    (
        "PHANTOM2",
        "Phantom (.5x) ___ _ _ ___ _ _ ___ _ _ ___ _ _ ___",
        [1.25, 0.625, -0.125, 0.125, -0.125, 0.125, -0.125],
    ),
    (
        "DASHDOT",
        "Dash dot __ . __ . __ . __ . __ . __ . __ . __",
        [1.4, 1.0, -0.2, 0.0, -0.2],
    ),
    (
        "DASHDOTX2",
        "Dash dot (2x) ____  .  ____  .  ____  .  ____",
        [2.4, 2.0, -0.2, 0.0, -0.2],
    ),
    (
        "DASHDOT2",
        "Dash dot (.5x) _ . _ . _ . _ . _ . _ . _ . _",
        [0.7, 0.5, -0.1, 0.0, -0.1],
    ),
    (
        "DOT",
        "Dot .  .  .  .  .  .  .  .  .  .  .  .  .  .  .  .",
        [0.2, 0.0, -0.2],
    ),
    (
        "DOTX2",
        "Dot (2x) .    .    .    .    .    .    .    . ",
        [0.4, 0.0, -0.4],
    ),
    (
        "DOT2",
        "Dot (.5) . . . . . . . . . . . . . . . . . . . ",
        [0.1, 0.0, -0.1],
    ),
    (
        "DIVIDE",
        "Divide __ . . __ . . __ . . __ . . __ . . __",
        [1.6, 1.0, -0.2, 0.0, -0.2, 0.0, -0.2],
    ),
    (
        "DIVIDEX2",
        "Divide (2x) ____  . .  ____  . .  ____  . .  ____",
        [2.6, 2.0, -0.2, 0.0, -0.2, 0.0, -0.2],
    ),
    (
        "DIVIDE2",
        "Divide(.5x) _ . _ . _ . _ . _ . _ . _ . _",
        [0.8, 0.5, -0.1, 0.0, -0.1, 0.0, -0.1],
    ),
]


def linetypes(scale: float = 1.0) -> List[LTypeDef]:
    """Creates a list of standard line types.
    Imperial units (in, ft, yd, ...) have a scale factor of 1.0, ISO units (m,
    cm, mm, ...) have a scale factor of 2.54, available as constant
    :attr:`ezdxf.tools.standards.ISO_LTYPE_FACTOR`.

    """
    return [scale_linetype(ltype, scale) for ltype in ANSI_LINE_TYPES]


def scale_linetype(ltype: LTypeDef, scale: float) -> LTypeDef:
    name, pattern_str, pattern = ltype
    return name, pattern_str, [round(e * scale, 3) for e in pattern]


def styles():
    """Creates a list of standard styles."""
    return [
        ("STANDARD", "txt"),
        ("OpenSans-Light", "OpenSans-Light.ttf"),
        ("OpenSans-Light-Italic", "OpenSans-LightItalic.ttf"),
        ("OpenSans", "OpenSans-Regular.ttf"),
        ("OpenSans-Italic", "OpenSans-Italic.ttf"),
        ("OpenSans-SemiBold", "OpenSans-SemiBold.ttf"),
        ("OpenSans-SemiBoldItalic", "OpenSans-SemiBoldItalic.ttf"),
        ("OpenSans-Bold", "OpenSans-Bold.ttf"),
        ("OpenSans-BoldItalic", "OpenSans-BoldItalic.ttf"),
        ("OpenSans-ExtraBold", "OpenSans-ExtraBold.ttf"),
        ("OpenSans-ExtraBoldItalic", "OpenSans-ExtraBoldItalic.ttf"),
        ("OpenSansCondensed-Bold", "OpenSansCondensed-Bold.ttf"),
        ("OpenSansCondensed-Light", "OpenSansCondensed-Light.ttf"),
        ("OpenSansCondensed-Italic", "OpenSansCondensed-LightItalic.ttf"),
        ("LiberationSans", "LiberationSans-Regular.ttf"),
        ("LiberationSans-Bold", "LiberationSans-Bold.ttf"),
        ("LiberationSans-BoldItalic", "LiberationSans-BoldItalic.ttf"),
        ("LiberationSans-Italic", "LiberationSans-Italic.ttf"),
        ("LiberationSerif", "LiberationSerif-Regular.ttf"),
        ("LiberationSerif-Bold", "LiberationSerif-Bold.ttf"),
        ("LiberationSerif-BoldItalic", "LiberationSerif-BoldItalic.ttf"),
        ("LiberationSerif-Italic", "LiberationSerif-Italic.ttf"),
        ("LiberationMono", "LiberationMono-Regular.ttf"),
        ("LiberationMono-Bold", "LiberationMono-Bold.ttf"),
        ("LiberationMono-BoldItalic", "LiberationMono-BoldItalic.ttf"),
        ("LiberationMono-Italic", "LiberationMono-Italic.ttf"),
    ]


VISUAL_STYLES = [
    {
        "description": "2dWireframe",
        "style_type": 4,
        "internal_use_only_flag": 0,
        "face_modifiers": 0,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "Basic",
        "style_type": 7,
        "internal_use_only_flag": 1,
        "face_modifiers": 1,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "Brighten",
        "style_type": 12,
        "internal_use_only_flag": 1,
        "face_modifiers": 2,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "ColorChange",
        "style_type": 16,
        "internal_use_only_flag": 1,
        "face_modifiers": 2,
        "face_opacity_level": 0.6,
        "color1": 8,
        "edge_hide_precision": 0,
    },
    {
        "description": "Conceptual",
        "style_type": 9,
        "internal_use_only_flag": 0,
        "face_modifiers": 3,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "Dim",
        "style_type": 11,
        "internal_use_only_flag": 1,
        "face_modifiers": 2,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "EdgeColorOff",
        "style_type": 22,
        "internal_use_only_flag": 1,
        "face_modifiers": 2,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "Facepattern",
        "style_type": 15,
        "internal_use_only_flag": 1,
        "face_modifiers": 2,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "Flat",
        "style_type": 0,
        "internal_use_only_flag": 1,
        "face_modifiers": 2,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "FlatWithEdges",
        "style_type": 1,
        "internal_use_only_flag": 1,
        "face_modifiers": 2,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "Gouraud",
        "style_type": 2,
        "internal_use_only_flag": 1,
        "face_modifiers": 2,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "GouraudWithEdges",
        "style_type": 3,
        "internal_use_only_flag": 1,
        "face_modifiers": 2,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "Hidden",
        "style_type": 6,
        "internal_use_only_flag": 0,
        "face_modifiers": 1,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "JitterOff",
        "style_type": 20,
        "internal_use_only_flag": 1,
        "face_modifiers": 2,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "Linepattern",
        "style_type": 14,
        "internal_use_only_flag": 1,
        "face_modifiers": 2,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "Modeling",
        "style_type": 10,
        "internal_use_only_flag": 0,
        "face_modifiers": 2,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "OverhangOff",
        "style_type": 21,
        "internal_use_only_flag": 1,
        "face_modifiers": 2,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "Realistic",
        "style_type": 8,
        "internal_use_only_flag": 0,
        "face_modifiers": 2,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "Shaded",
        "style_type": 27,
        "internal_use_only_flag": 0,
        "face_modifiers": 2,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "Shaded with edges",
        "style_type": 26,
        "internal_use_only_flag": 0,
        "face_modifiers": 2,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "Shades of Gray",
        "style_type": 23,
        "internal_use_only_flag": 0,
        "face_modifiers": 2,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "Sketchy",
        "style_type": 24,
        "internal_use_only_flag": 0,
        "face_modifiers": 1,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "Thicken",
        "style_type": 13,
        "internal_use_only_flag": 1,
        "face_modifiers": 2,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "Wireframe",
        "style_type": 5,
        "internal_use_only_flag": 0,
        "face_modifiers": 0,
        "face_opacity_level": 0.6,
        "color1": 7,
        "edge_hide_precision": 0,
    },
    {
        "description": "X-Ray",
        "style_type": 25,
        "internal_use_only_flag": 0,
        "face_modifiers": 2,
        "face_opacity_level": 0.5,
        "color1": 7,
        "edge_hide_precision": 0,
    },
]


def setup_visual_styles(doc: "Drawing"):
    objects = doc.objects
    vstyle_dict = doc.rootdict.get_required_dict("ACAD_VISUALSTYLE")
    vstyle_dict_handle = vstyle_dict.dxf.handle
    for vstyle in VISUAL_STYLES:
        vstyle["owner"] = vstyle_dict_handle
        vstyle_object = objects.add_dxf_object_with_reactor(
            "VISUALSTYLE", dxfattribs=vstyle
        )
        vstyle_dict[vstyle_object.dxf.description] = vstyle_object



# Copyright (c) 2019-2021, Manfred Moitzi
# License: MIT License
from typing import (
    TYPE_CHECKING,
    Iterable,
    Any,
    Union,
    List,
    Tuple,
    cast,
    Optional,
    Dict,
)
from ezdxf.math import Vec2
from ezdxf.entitydb import EntitySpace
from ezdxf.lldxf import const
from .base import BaseLayout

if TYPE_CHECKING:
    from ezdxf.eztypes import (
        GeoData,
        SortEntsTable,
        Vertex,
        Viewport,
        Drawing,
        DXFLayout,
        DXFGraphic,
        BlockLayout,
    )


def get_block_entity_space(
    doc: "Drawing", block_record_handle: str
) -> "EntitySpace":
    block_record = doc.entitydb[block_record_handle]
    return block_record.entity_space


class Layout(BaseLayout):
    """
    Layout representation - base class for :class:`Modelspace` and
    :class:`Paperspace`

    Every layout consist of a LAYOUT entity in the OBJECTS section, an
    associated BLOCK in the BLOCKS section and a BLOCK_RECORD_TABLE entry.

    layout_key: handle of the BLOCK_RECORD, every layout entity has this
        handle as owner attribute (entity.dxf.owner)

    There are 3 different layout types:

        1. :class:`Modelspace`
        2. active :class:`Paperspace` layout
        3. inactive :class:`Paperspace` layout

    Internal Structure

    For every layout exist a :class:`BlockLayout` object in
    :class:`BlocksSection` and a :class:`Layout` object
    (as :class:`Modelspace` or :class:`Paperspace`) in :class:`Layouts`.

    The entity space of the :class:`BlockLayout` object and the entity space
    of the :class:`Layout` object are the same object.

    """

    # plot_layout_flags of LAYOUT entity
    PLOT_VIEWPORT_BORDERS = 1
    SHOW_PLOT_STYLES = 2
    PLOT_CENTERED = 4
    PLOT_HIDDEN = 8
    USE_STANDARD_SCALE = 16
    PLOT_PLOTSTYLES = 32
    SCALE_LINEWEIGHTS = 64
    PRINT_LINEWEIGHTS = 128
    DRAW_VIEWPORTS_FIRST = 512
    MODEL_TYPE = 1024
    UPDATE_PAPER = 2048
    ZOOM_TO_PAPER_ON_UPDATE = 4096
    INITIALIZING = 8192
    PREV_PLOT_INIT = 16384

    def __init__(self, layout: "DXFLayout", doc: "Drawing"):
        self.dxf_layout = layout
        block_record = doc.entitydb[layout.dxf.block_record_handle]
        # link maybe broken
        block_record.dxf.layout = layout.dxf.handle
        super().__init__(block_record)

    @classmethod
    def new(
        cls, name: str, block_name: str, doc: "Drawing", dxfattribs: dict = None
    ) -> "Layout":
        """Returns the required structures for a new layout:

            - a :class:`BlockLayout` with BLOCK_RECORD, BLOCK and ENDBLK entities
            - LAYOUT entity in the objects section

        Args:
            name: layout name as shown in tabs of CAD applications e.g. 'Layout2'
            block_name: layout block name e.g. '*Paper_Space2'
            doc: drawing document
            dxfattribs: additional DXF attributes for LAYOUT entity

        (internal API)

        """
        block_layout = doc.blocks.new(block_name)  # type: BlockLayout
        dxfattribs = dxfattribs or {}
        dxfattribs.update(
            {
                "name": name,
                "block_record_handle": block_layout.block_record_handle,
            }
        )
        dxf_layout = doc.objects.new_entity("LAYOUT", dxfattribs=dxfattribs)
        return cls(dxf_layout, doc)  # type: ignore

    @classmethod
    def load(cls, layout: "DXFLayout", doc: "Drawing"):
        """Loading interface. (internal API)"""
        _layout = cls(layout, doc)
        _layout._repair_owner_tags()
        return _layout

    @property
    def name(self) -> str:
        """Layout name as shown in tabs of :term:`CAD` applications."""
        return self.dxf_layout.dxf.name

    @property  # dynamic DXF attribute dispatching, e.g. DXFLayout.dxf.layout_flags
    def dxf(self) -> Any:
        """Returns the DXF name space attribute of the associated
        :class:`~ezdxf.entities.DXFLayout` object.

        This enables direct access to the underlying LAYOUT entity,
        e.g. ``Layout.dxf.layout_flags``

        """
        return self.dxf_layout.dxf

    @property
    def block_record_name(self) -> str:
        """Returns the name of the associated BLOCK_RECORD as string."""
        return self.block_record.dxf.name

    def _repair_owner_tags(self) -> None:
        """Set `owner` and `paperspace` attributes of entities hosted by this
        layout to correct values.
        """
        layout_key = self.layout_key
        paperspace = 0 if self.is_modelspace else 1
        for entity in self:
            if entity.dxf.owner != layout_key:
                entity.dxf.owner = layout_key
            if entity.dxf.paperspace != paperspace:
                entity.dxf.paperspace = paperspace

    def __contains__(self, entity: Union["DXFGraphic", str]) -> bool:
        """Returns ``True`` if `entity` is stored in this layout.

        Args:
             entity: :class:`DXFGraphic` object or handle as hex string

        """
        if isinstance(entity, str):  # entity is a handle string
            entity = self.entitydb[entity]  # type: ignore
        return entity.dxf.owner == self.layout_key  # type: ignore

    def destroy(self) -> None:
        """Delete all entities and the layout itself from entity database and
        all linked structures.

        (internal API)
        """

        self.doc.objects.delete_entity(self.dxf_layout)
        # super() deletes block_record and associated entity space
        super().destroy()

    def get_sortents_table(self, create: bool = True) -> "SortEntsTable":
        """Get/Create the SORTENTSTABLE object associated to the layout.

        Args:
            create: new table if table do not exist and `create` is ``True``

        Raises:
            DXFValueError: if table not exist and `create` is ``False``

        (internal API)
        """
        xdict = self.get_extension_dict()
        try:
            sortents_table = xdict["ACAD_SORTENTS"]
        except const.DXFKeyError:
            if create:
                sortents_table = self.doc.objects.new_entity(
                    "SORTENTSTABLE",
                    dxfattribs={
                        "owner": xdict.handle,
                        "block_record_handle": self.layout_key,
                    },
                )
                xdict["ACAD_SORTENTS"] = sortents_table.dxf.handle
            else:
                raise const.DXFValueError(
                    "Extension dictionary entry ACAD_SORTENTS does not exist."
                )
        return sortents_table

    def set_redraw_order(
        self, handles: Union[Dict, Iterable[Tuple[str, str]]]
    ) -> None:
        """If the header variable $SORTENTS `Regen` flag (bit-code value 16)
        is set, AutoCAD regenerates entities in ascending handles order.

        To change redraw order associate a different sort handle to entities,
        this redefines the order in which the entities are regenerated.
        `handles` can be a dict of entity_handle and sort_handle as (k, v)
        pairs, or an iterable of (entity_handle, sort_handle) tuples.

        The sort_handle doesn't have to be unique, some or all entities can
        share the same sort handle and a sort handle can be an existing handle.

        The "0" handle can be used, but this sort_handle will be drawn as
        latest (on top of all other entities) and not as first as expected.

        Args:
            handles: iterable or dict of handle associations; an iterable
                of 2-tuples (entity_handle, sort_handle) or a dict (k, v)
                association as (entity_handle, sort_handle)

        """
        sortents = self.get_sortents_table()
        if isinstance(handles, dict):
            handles = handles.items()
        sortents.set_handles(handles)

    def get_redraw_order(self) -> Iterable[Tuple[str, str]]:
        """Returns iterable for all existing table entries as (entity_handle,
        sort_handle) pairs, see also :meth:`~Layout.set_redraw_order`.

        """
        if self.block_record.has_extension_dict:
            xdict = self.get_extension_dict()
        else:
            return tuple()

        try:
            sortents_table = xdict["ACAD_SORTENTS"]
        except const.DXFKeyError:
            return tuple()
        return iter(sortents_table)

    def reset_extents(
        self, extmin=(+1e20, +1e20, +1e20), extmax=(-1e20, -1e20, -1e20)
    ) -> None:
        """Reset `extents`_ to given values or the AutoCAD default values.

        "Drawing extents are the bounds of the area occupied by objects."
        (Quote Autodesk Knowledge Network)

        Args:
             extmin: minimum extents or (+1e20, +1e20, +1e20) as default value
             extmax: maximum extents or (-1e20, -1e20, -1e20) as default value

        """
        dxf = self.dxf_layout.dxf
        dxf.extmin = extmin
        dxf.extmax = extmax

    def reset_limits(self, limmin=None, limmax=None) -> None:
        """Reset `limits`_ to given values or the AutoCAD default values.

        "Sets an invisible rectangular boundary in the drawing area that can
        limit the grid display and limit clicking or entering point locations."
        (Quote Autodesk Knowledge Network)

        The :class:`Paperspace` class has an additional method
        :meth:`~Paperspace.reset_paper_limits` to deduce the default limits from
        the paper size settings.

        Args:
             extmin: minimum extents or (0, 0) as default
             extmax: maximum extents or (paper width, paper height) as default value

        """
        dxf = self.dxf_layout.dxf
        if limmin is None:
            limmin = (0, 0)
        if limmax is None:
            limmax = (dxf.paper_width, dxf.paper_height)

        dxf.limmin = limmin
        dxf.limmax = limmax

    def set_plot_type(self, value: int = 5) -> None:
        """
        === ============================================================
        0   last screen display
        1   drawing extents
        2   drawing limits
        3   view specific (defined by :attr:`Layout.dxf.plot_view_name`)
        4   window specific (defined by :meth:`Layout.set_plot_window_limits`)
        5   layout information (default)
        === ============================================================

        Args:
            value:  plot type

        Raises:
            DXFValueError: for `value` out of range

        """
        if 0 <= int(value) <= 5:
            self.dxf.plot_type = value
        else:
            raise const.DXFValueError("Plot type value out of range (0-5).")

    def set_plot_style(
        self, name: str = "ezdxf.ctb", show: bool = False
    ) -> None:
        """Set plot style file of type `.ctb`.

        Args:
            name: plot style filename
            show: show plot style effect in preview? (AutoCAD specific attribute)

        """
        self.dxf_layout.dxf.current_style_sheet = name
        self.use_plot_styles(True)
        self.show_plot_styles(show)

    def set_plot_window(
        self,
        lower_left: Tuple[float, float] = (0, 0),
        upper_right: Tuple[float, float] = (0, 0),
    ) -> None:
        """Set plot window size in (scaled) paper space units.

        Args:
            lower_left: lower left corner as 2D point
            upper_right: upper right corner as 2D point

        """
        x1, y1 = lower_left
        x2, y2 = upper_right
        dxf = self.dxf_layout.dxf
        dxf.plot_window_x1 = x1
        dxf.plot_window_y1 = y1
        dxf.plot_window_x2 = x2
        dxf.plot_window_y2 = y2
        self.set_plot_type(4)

    # plot layout flags setter
    def plot_viewport_borders(self, state: bool = True) -> None:
        self.set_plot_flags(self.PLOT_VIEWPORT_BORDERS, state)

    def show_plot_styles(self, state: bool = True) -> None:
        self.set_plot_flags(self.SHOW_PLOT_STYLES, state)

    def plot_centered(self, state: bool = True) -> None:
        self.set_plot_flags(self.PLOT_CENTERED, state)

    def plot_hidden(self, state: bool = True) -> None:
        self.set_plot_flags(self.PLOT_HIDDEN, state)

    def use_standard_scale(self, state: bool = True) -> None:
        self.set_plot_flags(self.USE_STANDARD_SCALE, state)

    def use_plot_styles(self, state: bool = True) -> None:
        self.set_plot_flags(self.PLOT_PLOTSTYLES, state)

    def scale_lineweights(self, state: bool = True) -> None:
        self.set_plot_flags(self.SCALE_LINEWEIGHTS, state)

    def print_lineweights(self, state: bool = True) -> None:
        self.set_plot_flags(self.PRINT_LINEWEIGHTS, state)

    def draw_viewports_first(self, state: bool = True) -> None:
        self.set_plot_flags(self.PRINT_LINEWEIGHTS, state)

    def model_type(self, state: bool = True) -> None:
        self.set_plot_flags(self.MODEL_TYPE, state)

    def update_paper(self, state: bool = True) -> None:
        self.set_plot_flags(self.UPDATE_PAPER, state)

    def zoom_to_paper_on_update(self, state: bool = True) -> None:
        self.set_plot_flags(self.ZOOM_TO_PAPER_ON_UPDATE, state)

    def plot_flags_initializing(self, state: bool = True) -> None:
        self.set_plot_flags(self.INITIALIZING, state)

    def prev_plot_init(self, state: bool = True) -> None:
        self.set_plot_flags(self.PREV_PLOT_INIT, state)

    def set_plot_flags(self, flag, state: bool = True) -> None:
        self.dxf_layout.set_flag_state(
            flag, state=state, name="plot_layout_flags"
        )


class Modelspace(Layout):
    """:class:`Modelspace` - not deletable, all entities of this layout are
    stored in the ENTITIES section of the DXF file, the associated
    "*Model_Space" block is empty, block name is fixed as "*Model_Space",
    the name is fixed as ""Model".

    """

    @property
    def name(self) -> str:
        """Name of modelspace is fixed as "Model"."""
        return "Model"

    def new_geodata(self, dxfattribs: dict = None) -> "GeoData":
        """Creates a new :class:`GeoData` entity and replaces existing ones.
        The GEODATA entity resides in the OBJECTS section and not in the
        modelspace, it is linked to the modelspace by an
        :class:`~ezdxf.entities.ExtensionDict` located in BLOCK_RECORD of the
        modelspace.

        The GEODATA entity requires DXF R2010. The DXF reference does not
        document if other layouts than the modelspace supports geo referencing,
        so I assume getting/setting geo data may only make sense for the
        modelspace.

        Args:
            dxfattribs: DXF attributes for :class:`~ezdxf.entities.GeoData` entity

        """
        if self.doc.dxfversion < const.DXF2010:
            raise const.DXFValueError(
                "GEODATA entity requires DXF R2010 or later."
            )

        if dxfattribs is None:
            dxfattribs = {}
        xdict = self.get_extension_dict()
        geodata = self.doc.objects.add_geodata(
            owner=xdict.dictionary.dxf.handle,
            dxfattribs=dxfattribs,
        )
        xdict["ACAD_GEOGRAPHICDATA"] = geodata
        return geodata

    def get_geodata(self) -> Optional["GeoData"]:
        """Returns the :class:`~ezdxf.entities.GeoData` entity associated to
        the modelspace or ``None``.
        """
        try:
            xdict = self.block_record.get_extension_dict()
        except AttributeError:
            return None
        try:
            return xdict["ACAD_GEOGRAPHICDATA"]
        except const.DXFKeyError:
            return None


class Paperspace(Layout):
    """There are two kind of paperspace layouts:

    1. Active Layout - all entities of this layout are stored in the ENTITIES
       section, the associated "*Paper_Space" block is empty, block name
       "*Paper_Space" is mandatory and also marks the active layout, the layout
       name can be an arbitrary string.

    2. Inactive Layout - all entities of this layouts are stored in the
       associated BLOCK called "*Paper_SpaceN", where "N" is an arbitrary
       number, I don't know if the block name schema "*Paper_SpaceN" is
       mandatory, the layout name can be an arbitrary string.

    There is no different handling for active layouts and inactive layouts in
    `ezdxf`, this differentiation is just for AutoCAD important and it is not
    documented in the DXF reference.

    """

    def rename(self, name: str) -> None:
        """Rename layout to `name`, changes the name displayed in tabs by
        CAD applications, not the internal BLOCK name. (internal API)

        Use method :meth:`~ezdxf.layouts.Layouts.rename` of the
        :meth:`~ezdxf.layouts.Layouts` class to rename paper space
        layouts.

        """
        self.dxf_layout.dxf.name = name

    def viewports(self) -> List["Viewport"]:
        """Get all VIEWPORT entities defined in this paperspace layout.
        Returns a list of :class:`~ezdxf.entities.Viewport` objects, sorted by
        id, the first entity is always the main viewport with an id of 1.

        """
        vports = [entity for entity in self if entity.dxftype() == "VIEWPORT"]
        vports.sort(key=lambda e: e.dxf.id)
        return vports  # type: ignore

    def main_viewport(self) -> Optional["Viewport"]:
        """Returns the main viewport of this paper space layout, or ``None``
        if no main viewport exist.

        """
        viewports = self.viewports()
        if len(viewports):
            return viewports[0]
        else:
            return None

    def renumber_viewports(self) -> None:
        """Reassign viewport ids. (internal API)"""
        for num, viewport in enumerate(self.viewports(), start=1):
            viewport.dxf.id = num

    def add_viewport(
        self,
        center: "Vertex",
        size: Tuple[float, float],
        view_center_point: "Vertex",
        view_height: float,
        dxfattribs: dict = None,
    ) -> "Viewport":
        """Add a new :class:`~ezdxf.entities.Viewport` entity."""
        dxfattribs = dxfattribs or {}
        width, height = size
        attribs = {
            "center": center,
            "width": width,
            "height": height,
            "status": 1,  # by default highest priority (stack order)
            "layer": "VIEWPORTS",
            # use separated layer to turn off for plotting
            "view_center_point": view_center_point,
            "view_height": view_height,
        }
        attribs.update(dxfattribs)
        viewport = cast("Viewport", self.new_entity("VIEWPORT", attribs))
        viewport.dxf.id = viewport.get_next_viewport_id()
        return viewport

    def reset_viewports(self) -> None:
        """Delete all existing viewports, and create a new main viewport."""
        # remove existing viewports
        for viewport in self.viewports():
            self.delete_entity(viewport)
        self.add_new_main_viewport()

    def reset_main_viewport(
        self, center: "Vertex" = None, size: "Vertex" = None
    ) -> "Viewport":
        """Reset the main viewport of this paper space layout to the given
        values, or reset them to the default values, deduced from the paper
        settings. Creates a new main viewport if none exist.

        Ezdxf does not create a main viewport by default, because CAD
        applications don't require one.

        Args:
            center: center of the viewport in paper space units
            size: viewport size as (width, height) tuple in paper space units

        """
        viewports = self.viewports()
        if len(viewports) == 0 or viewports[0].dxf.id != 1:
            viewport = self.add_new_main_viewport()
        else:
            viewport = viewports[0]
        default_center, default_size = self.default_viewport_config()
        if center is None:
            center = default_center
        if size is None:
            size = default_size

        viewport.dxf.center = center
        width, height = size
        viewport.dxf.width = width
        viewport.dxf.height = height
        return viewport

    def default_viewport_config(
        self,
    ) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        dxf = self.dxf_layout.dxf
        if dxf.plot_paper_units == 0:  # inches
            unit_factor = 25.4
        else:  # mm
            unit_factor = 1.0

        # all paper parameters in mm!
        # all viewport parameters in paper space units inch/mm + scale factor!
        scale_factor = dxf.scale_denominator / dxf.scale_numerator

        def paper_units(value):
            return value / unit_factor * scale_factor

        paper_width = paper_units(dxf.paper_width)
        paper_height = paper_units(dxf.paper_height)
        # plot origin offset
        x_offset = paper_units(dxf.plot_origin_x_offset)
        y_offset = paper_units(dxf.plot_origin_y_offset)

        # printing area
        printable_width = (
            paper_width
            - paper_units(dxf.left_margin)
            - paper_units(dxf.right_margin)
        )
        printable_height = (
            paper_height
            - paper_units(dxf.bottom_margin)
            - paper_units(dxf.top_margin)
        )

        # AutoCAD viewport (window) size
        vp_width = paper_width * 1.1
        vp_height = paper_height * 1.1

        # center of printing area
        center = (
            printable_width / 2 - x_offset,
            printable_height / 2 - y_offset,
        )
        return center, (vp_width, vp_height)

    def add_new_main_viewport(self) -> "Viewport":
        """Add a new main viewport."""
        center, size = self.default_viewport_config()
        vp_height = size[1]
        # create 'main' viewport
        main_viewport = self.add_viewport(
            center=center,  # no influence to 'main' viewport?
            size=size,  # I don't get it, just use paper size!
            view_center_point=center,  # same as center
            view_height=vp_height,  # view height in paper space units
        )
        main_viewport.dxf.id = 1  # set as main viewport
        main_viewport.dxf.flags = 557088  # AutoCAD default value
        self.dxf_layout.dxf.viewport_handle = main_viewport.dxf.handle
        return main_viewport

    def page_setup(
        self,
        size: Tuple[float, float] = (297, 210),
        margins: Tuple[float, float, float, float] = (10, 15, 10, 15),
        units: str = "mm",
        offset: Tuple[float, float] = (0, 0),
        rotation: int = 0,
        scale: Union[int, Tuple[float, float]] = 16,
        name: str = "ezdxf",
        device: str = "DWG to PDF.pc3",
    ) -> None:
        """Setup plot settings and paper size and reset viewports.
        All parameters in given `units` (mm or inch).

        Reset paper limits, extents and viewports.

        Args:
            size: paper size as (width, height) tuple
            margins: (top, right, bottom, left) hint: clockwise
            units: "mm" or "inch"
            offset: plot origin offset is 2D point
            rotation: see table Rotation
            scale: integer in range [0, 32] defines a standard scale type or
                as tuple(numerator, denominator) e.g. (1, 50) for scale 1:50
            name: paper name prefix "{name}_({width}_x_{height}_{unit})"
            device: device .pc3 configuration file or system printer name

        === ============
        int Rotation
        === ============
        0   no rotation
        1   90 degrees counter-clockwise
        2   upside-down
        3   90 degrees clockwise
        === ============

        """
        if int(rotation) not in (0, 1, 2, 3):
            raise const.DXFValueError("valid rotation values: 0-3")

        if isinstance(scale, tuple):
            standard_scale = 16
            scale_num, scale_denom = scale
        elif isinstance(scale, int):
            standard_scale = scale
            scale_num, scale_denom = const.STD_SCALES.get(
                standard_scale, (1.0, 1.0)
            )
        else:
            raise const.DXFTypeError(
                "Scale has to be an int or a tuple(numerator, denominator)"
            )
        if scale_num == 0:
            raise const.DXFValueError("Scale numerator can't be 0.")
        if scale_denom == 0:
            raise const.DXFValueError("Scale denominator can't be 0.")

        self.use_standard_scale(False)  # works best, don't know why
        paper_width, paper_height = size
        margin_top, margin_right, margin_bottom, margin_left = margins
        units = units.lower()
        if units.startswith("inch"):
            units = "Inches"
            plot_paper_units = 0
            unit_factor = 25.4  # inch to mm
        elif units == "mm":
            units = "MM"
            plot_paper_units = 1
            unit_factor = 1.0
        else:
            raise const.DXFValueError('Supported units: "mm" and "inch"')

        # Setup PLOTSETTINGS
        # all paper sizes in mm
        dxf = self.dxf_layout.dxf
        dxf.page_setup_name = ""
        dxf.plot_configuration_file = device
        dxf.paper_size = (
            f"{name}_({paper_width:.2f}_x_{paper_height:.2f}_{units})"
        )
        dxf.left_margin = margin_left * unit_factor
        dxf.bottom_margin = margin_bottom * unit_factor
        dxf.right_margin = margin_right * unit_factor
        dxf.top_margin = margin_top * unit_factor
        dxf.paper_width = paper_width * unit_factor
        dxf.paper_height = paper_height * unit_factor
        dxf.scale_numerator = scale_num
        dxf.scale_denominator = scale_denom
        dxf.plot_paper_units = plot_paper_units
        dxf.plot_rotation = rotation

        x_offset, y_offset = offset
        dxf.plot_origin_x_offset = x_offset * unit_factor  # conversion to mm
        dxf.plot_origin_y_offset = y_offset * unit_factor  # conversion to mm
        dxf.standard_scale_type = standard_scale
        dxf.unit_factor = 1.0 / unit_factor  # 1/1 for mm; 1/25.4 ... for inch

        # Setup Layout
        self.reset_paper_limits()
        self.reset_extents()
        self.reset_viewports()

    def reset_paper_limits(self) -> None:
        """Set paper limits to default values, all values in paperspace units
        but without plot scale (?).

        """
        dxf = self.dxf_layout.dxf
        if dxf.plot_paper_units == 0:  # inch
            unit_factor = 25.4
        else:  # mm
            unit_factor = 1.0

        # all paper sizes are stored in mm
        paper_width = dxf.paper_width / unit_factor  # in plot paper units
        paper_height = dxf.paper_height / unit_factor  # in plot paper units
        left_margin = dxf.left_margin / unit_factor
        bottom_margin = dxf.bottom_margin / unit_factor
        x_offset = dxf.plot_origin_x_offset / unit_factor
        y_offset = dxf.plot_origin_y_offset / unit_factor
        # plot origin is the lower left corner of the printable paper area
        # limits are the paper borders relative to the plot origin
        shift_x = left_margin + x_offset
        shift_y = bottom_margin + y_offset
        dxf.limmin = (-shift_x, -shift_y)  # paper space units
        dxf.limmax = (paper_width - shift_x, paper_height - shift_y)

    def get_paper_limits(self) -> Tuple[Vec2, Vec2]:
        """Returns paper limits in plot paper units, relative to the plot origin.

        plot origin = lower left corner of printable area + plot origin offset

        Returns:
            tuple (Vec2(x1, y1), Vec2(x2, y2)), lower left corner is (x1, y1),
            upper right corner is (x2, y2).

        """
        return Vec2(self.dxf.limmin), Vec2(self.dxf.limmax)

    def page_setup_r12(
        self,
        size: Tuple[int, int] = (297, 210),
        margins: Tuple[int, int, int, int] = (0, 0, 0, 0),
        units: str = "mm",
        offset: Tuple[int, int] = (0, 0),
        rotation: float = 0,
        scale: Union[int, Tuple[float, float]] = 16,
    ) -> None:

        # remove existing viewports
        for viewport in self.viewports():
            self.delete_entity(viewport)

        if int(rotation) not in (0, 1, 2, 3):
            raise const.DXFValueError("Valid rotation values: 0-3")

        if isinstance(scale, int):
            scale_num, scale_denom = const.STD_SCALES.get(scale, (1, 1))
        else:
            scale_num, scale_denom = scale

        if scale_num == 0:
            raise const.DXFValueError("Scale numerator can't be 0.")
        if scale_denom == 0:
            raise const.DXFValueError("Scale denominator can't be 0.")

        scale_factor = scale_denom / scale_num

        # TODO: don't know how to set inch or mm mode in R12
        units = units.lower()
        if units.startswith("inch"):
            units = "Inches"
            plot_paper_units = 0
            unit_factor = 25.4  # inch to mm
        elif units == "mm":
            units = "MM"
            plot_paper_units = 1
            unit_factor = 1.0
        else:
            raise const.DXFValueError('Supported units: "mm" and "inch"')

        # all viewport parameters are scaled paper space units
        def paper_units(value):
            return value * scale_factor

        # TODO: don't know how paper setup in DXF R12 works
        paper_width, paper_height = size

        # TODO: don't know how margins setup in DXF R12 works
        margin_top, margin_right, margin_bottom, margin_left = margins

        paper_width = paper_units(size[0])
        paper_height = paper_units(size[1])

        plimmin = self.doc.header["$PLIMMIN"] = (0, 0)
        plimmax = self.doc.header["$PLIMMAX"] = (paper_width, paper_height)

        # TODO: don't know how paper setup in DXF R12 works

        pextmin = self.doc.header["$PEXTMIN"] = (0, 0, 0)
        pextmax = self.doc.header["$PEXTMAX"] = (paper_width, paper_height, 0)

        # printing area
        printable_width = (
            paper_width - paper_units(margin_left) - paper_units(margin_right)
        )
        printable_height = (
            paper_height - paper_units(margin_bottom) - paper_units(margin_top)
        )

        # AutoCAD viewport (window) size
        vp_width = paper_width * 1.1
        vp_height = paper_height * 1.1

        # center of printing area
        center = (printable_width / 2, printable_height / 2)

        # create 'main' viewport
        main_viewport = self.add_viewport(
            center=center,  # no influence to 'main' viewport?
            size=(vp_width, vp_height),  # I don't get it, just use paper size!
            view_center_point=center,  # same as center
            view_height=vp_height,  # view height in paper space units
        )
        main_viewport.dxf.id = 1  # set as main viewport
        main_viewport.dxf.status = 2  # AutoCAD default value
        main_viewport.dxf.render_mode = 1000  # AutoDesk default (view mode?)

    def get_paper_limits_r12(self) -> Tuple[Vec2, Vec2]:
        """Returns paper limits in plot paper units."""
        limmin = self.doc.header.get("$PLIMMIN", (0, 0))
        limmax = self.doc.header.get("$PLIMMAX", (0, 0))
        return Vec2(limmin), Vec2(limmax)



# Copyright (c) 2019-2021, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterable
from ezdxf.math import Vec2
from ezdxf.math.line import ConstructionRay, ParallelRaysError

if TYPE_CHECKING:
    from ezdxf.eztypes import Vertex

__all__ = ["offset_vertices_2d"]


def offset_vertices_2d(
    vertices: Iterable["Vertex"], offset: float, closed: bool = False
) -> Iterable["Vec2"]:
    """Yields vertices of the offset line to the shape defined by `vertices`.
    The source shape consist of straight segments and is located in the xy-plane,
    the z-axis of input vertices is ignored. Takes closed shapes into account if
    argument `closed` is ``True``, which yields intersection of first and last
    offset segment as first vertex for a closed shape. For closed shapes the
    first and last vertex can be equal, else an implicit closing segment from
    last to first vertex is added. A shape  with equal first and last vertex is
    not handled automatically as closed shape.

    .. warning::

        Adjacent collinear segments in `opposite` directions, same as a turn by
        180 degree (U-turn), leads to unexpected results.

    Args:
        vertices: source shape defined by vertices
        offset: line offset perpendicular to direction of shape segments defined
            by vertices order, offset > ``0`` is 'left' of line segment,
            offset < ``0`` is 'right' of line segment
        closed: ``True`` to handle as closed shape

    """
    _vertices = Vec2.list(vertices)
    if len(_vertices) < 2:
        raise ValueError("2 or more vertices required.")

    if closed and not _vertices[0].isclose(_vertices[-1]):
        # append first vertex as last vertex to close shape
        _vertices.append(_vertices[0])

    # create offset segments
    offset_segments = list()
    for start, end in zip(_vertices[:-1], _vertices[1:]):
        offset_vec = (end - start).orthogonal().normalize(offset)
        offset_segments.append((start + offset_vec, end + offset_vec))

    if closed:  # insert last segment also as first segment
        offset_segments.insert(0, offset_segments[-1])

    # first offset vertex = start point of first segment for open shapes
    if not closed:
        yield offset_segments[0][0]

    # yield intersection points of offset_segments
    if len(offset_segments) > 1:
        for (start1, end1), (start2, end2) in zip(
            offset_segments[:-1], offset_segments[1:]
        ):
            try:  # the usual case
                yield ConstructionRay(start1, end1).intersect(
                    ConstructionRay(start2, end2)
                )
            except ParallelRaysError:  # collinear segments
                yield end1
                if not end1.isclose(start2):  # it's an U-turn (180 deg)
                    # creates an additional vertex!
                    yield start2

    # last offset vertex = end point of last segment for open shapes
    if not closed:
        yield offset_segments[-1][1]
